from io import BytesIO
from typing import Dict

import pandas as pd
from wbcore.contrib.io.models import ImportSource

from wbportfolio.models import Product, Trade


def parse_row(obj: Dict, import_source: ImportSource) -> Dict:
    if _id := obj["underlying_instrument_id"]:
        product = Product.objects.get(id=_id)
        shares = -1 * obj["shares"]
        portfolio = product.primary_portfolio
        return {
            "underlying_instrument": {"id": product.id, "instrument_type": "product"},
            "currency__key": product.currency.key,
            "portfolio": portfolio.id,
            "transaction_date": obj["transaction_date"].strftime("%Y-%m-%d"),
            "shares": shares,
            "bank": obj["custodian"] if obj["custodian"] else "NA",
            "transaction_subtype": Trade.Type.REDEMPTION if shares < 0 else Trade.Type.SUBSCRIPTION,
            "price": round(obj["price"], 6),
        }


def parse(import_source):
    def _get_underlying_instrument_id(isin):
        try:
            product = Product.objects.get(isin=isin)
            return product.id
        except Product.DoesNotExist:
            import_source.log += f"Product with ISIN {isin} does not exists."

            return None

    df_dict = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl")
    df_dict = df_dict.rename(
        columns={
            "Trade Date": "transaction_date",
            "ISIN": "underlying_instrument__isin",
            "Client Side": "client_side",
            "Net Quantity": "shares",
            "Price": "price",
            "Custodian": "custodian",
        }
    )
    df_dict = df_dict.where(pd.notnull(df_dict), None)
    df_dict["transaction_date"] = pd.to_datetime(df_dict["transaction_date"])
    df_dict["underlying_instrument__isin"] = df_dict["underlying_instrument__isin"].str.strip()
    df_dict["underlying_instrument_id"] = df_dict["underlying_instrument__isin"].apply(
        lambda x: _get_underlying_instrument_id(x)
    )

    data = list()
    for d in df_dict.to_dict("records"):
        parsed_row = parse_row(d, import_source)
        if parsed_row:
            data.append(parse_row(d, import_source))
    underlying_instruments = [int(_id) for _id in df_dict.underlying_instrument_id.astype("int").unique()]

    return {
        "data": data,
        "history": {
            "underlying_instruments": underlying_instruments,
            "transaction_date": df_dict.transaction_date.max().strftime("%Y-%m-%d"),
        },
    }



from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbfdm.serializers import InvestableInstrumentRepresentationSerializer
from wbfdm.serializers.exchanges import ExchangeRepresentationSerializer

from wbportfolio.models import AssetPosition, Portfolio

from .portfolios import PortfolioRepresentationSerializer


class AssetPositionModelSerializer(wb_serializers.ModelSerializer):
    underlying_quote_isin = wb_serializers.CharField(read_only=True)
    underlying_quote_ticker = wb_serializers.CharField(read_only=True)
    underlying_quote_name = wb_serializers.CharField(read_only=True)
    _underlying_quote = InvestableInstrumentRepresentationSerializer(source="underlying_quote", label_key="name")
    _currency = CurrencyRepresentationSerializer(source="currency")
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    _exchange = ExchangeRepresentationSerializer(source="exchange")
    _portfolio_created = PortfolioRepresentationSerializer(source="portfolio_created")

    shares = wb_serializers.DecimalField(read_only=True, decimal_places=4, max_digits=14)
    total_value_fx_usd = wb_serializers.FloatField(read_only=True, precision=2)
    liquidity = wb_serializers.FloatField(read_only=True, default=0, precision=2)
    market_share = wb_serializers.FloatField(read_only=True, precision=6)
    market_capitalization = wb_serializers.FloatField(read_only=True)
    price = wb_serializers.DecimalField(read_only=True, decimal_places=4, max_digits=14)
    currency_fx_rate = wb_serializers.FloatField(read_only=True)
    price_fx_portfolio = wb_serializers.FloatField(read_only=True)
    total_value = wb_serializers.FloatField(read_only=True)
    total_value_fx_portfolio = wb_serializers.FloatField(read_only=True)

    is_invested = wb_serializers.BooleanField(read_only=True, default=False)
    currency_symbol = wb_serializers.CharField(read_only=True)
    portfolio_currency_symbol = wb_serializers.CharField(read_only=True)

    class Meta:
        decorators = {
            "total_value_fx_portfolio": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{portfolio_currency_symbol}}"
            ),
            "total_value": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
            "price": wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}"),
            "total_value_fx_usd": wb_serializers.decorator(decorator_type="text", position="left", value="$"),
        }
        percent_fields = ["weighting", "market_share"]
        model = AssetPosition
        fields = (
            "id",
            "asset_valuation_date",
            "exchange",
            "price_denotation",
            "date",
            "shares",
            "weighting",
            "portfolio",
            "portfolio_created",
            "currency",
            "_currency",
            "currency_fx_rate",
            "market_share",
            "market_capitalization",
            "currency_fx_rate",
            "price",
            "price_fx_portfolio",
            "total_value",
            "total_value_fx_usd",
            "total_value_fx_portfolio",
            "liquidity",
            "is_estimated",
            "underlying_quote_name",
            "underlying_quote_isin",
            "underlying_quote_ticker",
            "underlying_quote",
            "_underlying_quote",
            "_portfolio",
            "_portfolio_created",
            "_exchange",
            "exchange",
            "is_invested",
            "_additional_resources",
            "currency_symbol",
            "portfolio_currency_symbol",
        )


class AssetPositionPortfolioModelSerializer(AssetPositionModelSerializer):
    _portfolio = None

    class Meta(AssetPositionModelSerializer.Meta):
        fields = (
            "id",
            "is_estimated",
            "underlying_quote_name",
            "underlying_quote_isin",
            "underlying_quote_ticker",
            "underlying_quote",
            "_underlying_quote",
            "currency",
            "_currency",
            "exchange",
            "_exchange",
            "price",
            "shares",
            "total_value",
            "currency_fx_rate",
            "total_value_fx_portfolio",
            "total_value_fx_usd",
            "weighting",
            "portfolio_created",
            "_portfolio_created",
            "market_share",
            "asset_valuation_date",
            "liquidity",
            "currency_symbol",
            "portfolio_currency_symbol",
        )
        read_only_fields = fields


class AssetPositionAggregatedPortfolioModelSerializer(AssetPositionPortfolioModelSerializer):
    # If the aggregated serializer is used, we only show the related model representation for performance
    currency = wb_serializers.CharField()

    class Meta(AssetPositionModelSerializer.Meta):
        fields = (
            "id",
            "is_estimated",
            "underlying_quote_isin",
            "underlying_quote_ticker",
            "underlying_quote_name",
            "currency",
            "price",
            "shares",
            "total_value",
            "currency_fx_rate",
            "total_value_fx_portfolio",
            "total_value_fx_usd",
            "weighting",
            "market_share",
            "asset_valuation_date",
            "liquidity",
            "currency_symbol",
            "portfolio_currency_symbol",
        )
        read_only_fields = fields


class AssetPositionInstrumentModelSerializer(AssetPositionModelSerializer):
    class Meta(AssetPositionModelSerializer.Meta):
        fields = AssetPositionModelSerializer.Meta.fields


class CashPositionPortfolioModelSerializer(wb_serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    sum_total_value_fx_usd = wb_serializers.FloatField(read_only=True)
    portfolio = wb_serializers.PrimaryKeyRelatedField(queryset=Portfolio.objects.all())

    class Meta:
        decorators = {
            "sum_total_value_fx_usd": wb_serializers.decorator(decorator_type="text", position="left", value="$")
        }
        model = AssetPosition
        fields = ("id", "portfolio", "_portfolio", "sum_total_value_fx_usd", "date", "_additional_resources")



import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from psycopg.types.range import NumericRange
from wbcompliance.factories.risk_management import RuleThresholdFactory

from wbportfolio.risk_management.backends.product_integrity import (
    RuleBackend as ProductRuleBackend,
)

fake = Faker()


@pytest.mark.django_db
class TestProductRuleModel:
    @pytest.fixture
    def product_backend(
        self,
        weekday,
        product,
    ):
        return ProductRuleBackend(
            weekday,
            product,
            thresholds=[
                RuleThresholdFactory.create(range=NumericRange(lower=1, upper=2), severity__name="LOW"),
                RuleThresholdFactory.create(range=NumericRange(lower=2, upper=None), severity__name="HIGH"),
            ],  # detect any -20% perf
        )

    def test_check_rule_product_data_integrity(
        self, weekday, product, instrument_price_factory, asset_position_factory, product_backend
    ):
        asset_position_factory.create(date=weekday - BDay(2), portfolio=product.portfolio, is_estimated=False)
        instrument_price_factory.create(date=weekday - BDay(1), instrument=product, calculated=False)

        res = list(product_backend.check_rule())
        assert len(res) == 1
        incident = res[0]
        assert incident.severity.name == "HIGH"
        assert incident.breached_value == str(2)
        assert incident.report_details["Lag"] == 2
        assert incident.report_details["Last Asset Position Datapoint"] == f"{(weekday - BDay(2)).date():%d.%m.%Y}"
        assert incident.report_details["Last Valuation Datapoint"] == f"{(weekday - BDay(1)).date():%d.%m.%Y}"

        instrument_price_factory.create(date=weekday, instrument=product, calculated=False)
        asset_position_factory.create(date=weekday, portfolio=product.portfolio, is_estimated=False)
        res = list(product_backend.check_rule())
        assert len(res) == 0



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_manager

TRANSACTION_MENUITEM = MenuItem(
    label="Transactions",
    endpoint="wbportfolio:transaction-list",
    permission=ItemPermission(method=is_manager, permissions=["wbportfolio.view_transaction"]),
)



import factory
from wbfdm.factories.instruments import InstrumentFactory

from wbportfolio.models import ProductGroup, ProductGroupRepresentant


class ProductGroupFactory(InstrumentFactory):
    class Meta:
        model = ProductGroup

    type = factory.Faker("random_element", elements=[x[0] for x in ProductGroup.ProductGroupType.choices])
    category = factory.Faker("random_element", elements=[x[0] for x in ProductGroup.ProductGroupCategory.choices])
    umbrella = factory.Faker("company")
    management_company = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")
    depositary = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")
    transfer_agent = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")
    administrator = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")
    investment_manager = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")
    auditor = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")
    paying_agent = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")


class ProductGroupRepresentantFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ProductGroupRepresentant

    product_group = factory.SubFactory(ProductGroupFactory)
    representant = factory.SubFactory("wbcore.contrib.directory.factories.entries.CompanyFactory")
    country = factory.SubFactory("wbcore.contrib.geography.factories.CountryFactory")



from datetime import date

from django.db.models import Exists, OuterRef
from django.dispatch import receiver
from wbcore import filters as wb_filters
from wbcore.signals.filters import add_filters
from wbfdm.filters import BaseClassifiedInstrumentFilterSet, ClassificationFilter
from wbfdm.models import InstrumentClassificationThroughModel

from wbportfolio.models import AssetPosition, Portfolio


@receiver(add_filters, sender=ClassificationFilter)
def add_portfolio_filter(sender, request=None, *args, **kwargs):
    def _filter_portfolio(queryset, name, value):
        if value:
            try:
                last_position_date = value.assets.latest("date").date
            except AssetPosition.DoesNotExist:
                last_position_date = (
                    AssetPosition.objects.latest("date").date if AssetPosition.objects.exists() else date.today()
                )
            invested_instruments = AssetPosition.get_invested_instruments(last_position_date, portfolio=value)
            rels = InstrumentClassificationThroughModel.objects.filter(
                instrument__in=invested_instruments.values("root")
            )
            return queryset.filter(id__in=rels.values("classification"))
        return queryset

    def _filter_invested(queryset, name, value):
        if value:
            last_position_date = (
                AssetPosition.objects.latest("date").date if AssetPosition.objects.exists() else date.today()
            )
            invested_instruments = AssetPosition.get_invested_instruments(last_position_date)
            rels = InstrumentClassificationThroughModel.objects.filter(
                instrument__in=invested_instruments.values("root")
            )
            return queryset.filter(id__in=rels.values("classification"))
        return queryset

    return {
        "portfolio": wb_filters.ModelChoiceFilter(
            label="Associated Portfolio",
            queryset=Portfolio.objects.all(),
            endpoint=Portfolio.get_representation_endpoint(),
            value_key=Portfolio.get_representation_value_key(),
            label_key=Portfolio.get_representation_label_key(),
            method=_filter_portfolio,
        ),
        "only_invested": wb_filters.BooleanFilter(
            label="Only invested instruments (anytime)",
            method=_filter_invested,
        ),
    }


@receiver(add_filters, sender=BaseClassifiedInstrumentFilterSet)
def add_classification_instrument_filter(sender, request=None, *args, **kwargs):
    def _filter_invested(queryset, name, value):
        if value:
            last_invested_date = (
                AssetPosition.objects.latest("date").date if AssetPosition.objects.exists() else date.today()
            )
            invested_instruments = AssetPosition.get_invested_instruments(last_invested_date)
            return queryset.filter(Exists(invested_instruments.filter(root=OuterRef("instrument"))))
        return queryset

    return {
        "only_invested": wb_filters.BooleanFilter(
            method=_filter_invested, label="Invested Instruments (last date)", default=False
        )
    }



import json
from typing import TYPE_CHECKING

from django.db.models import F, FloatField, Sum
from django.db.models.functions import Cast, ExtractYear
from langchain_core.messages import HumanMessage, SystemMessage
from wbfdm.models import Instrument

if TYPE_CHECKING:
    from langchain_core.messages import BaseMessage
    from wbcrm.models import Account


def get_holding_prompt(account: "Account") -> list["BaseMessage"]:
    from wbportfolio.models import Product
    from wbportfolio.models.transactions.claim import Claim

    products = (
        Claim.get_valid_and_approved_claims(account=account)
        .distinct("product")
        .values_list("product", "product__isin")
    )

    performances = {}
    for product_id, product_name in products:
        performances[product_name] = Instrument.extract_annual_performance_df(
            Product.objects.get(id=product_id).get_prices_df()
        ).to_dict()

    return [
        SystemMessage(
            "The following products are held by the account holder. Analyze their performances and check correlations between the holdings and their performances/interactions."
        ),
        HumanMessage(json.dumps(performances)),
    ]


def get_performances_prompt(account: "Account") -> list["BaseMessage"]:
    from wbportfolio.models.transactions.claim import Claim

    holdings = (
        Claim.get_valid_and_approved_claims(account=account)
        .annotate(year=ExtractYear("date"))
        .values("year", "product")
        .annotate(
            sum_shares=Cast(Sum("shares"), FloatField()),
            product_name=F("product__name"),
            product_isin=F("product__isin"),
        )
        .values("product_name", "product_isin", "sum_shares", "year")
    )

    return [
        SystemMessage(
            "The following holdings (subscriptions/redemptions) have been found for this account. Please include this data in the analysis and check if there is any correlation between the holding data and the interactions."
        ),
        HumanMessage(json.dumps(list(holdings.order_by("year", "product")))),
    ]



from wbportfolio.models import PortfolioRole


def is_manager(request):
    return PortfolioRole.is_manager(request.user.profile)


def is_portfolio_manager(request):
    return PortfolioRole.is_portfolio_manager(request.user.profile)


def is_analyst(request):
    return PortfolioRole.is_analyst(request.user.profile)



import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from rest_framework.test import APIRequestFactory

from wbportfolio.viewsets.positions import AggregatedAssetPositionLiquidityPandasView

fake = Faker()


@pytest.mark.django_db
class TestPandasViewSet:
    def test_get_dataframe(self, instrument_price_factory, asset_position_factory, equity, portfolio, weekday):
        request = APIRequestFactory().get("")
        request.query_params = dict()
        # Make instances of AP, instrument prices.
        i1 = instrument_price_factory.create(
            instrument=equity,
            date=weekday,
            volume_50d=1000,
            net_value=10,
            instrument__currency__key="USD",
        )
        i2 = instrument_price_factory.create(
            instrument=equity,
            date=(i1.date - BDay(1)).date(),
            volume_50d=1100,
            net_value=11,
            instrument__currency__key="USD",
        )
        a1 = asset_position_factory.create(
            portfolio=portfolio,
            portfolio__currency__key="USD",
            underlying_instrument=i1.instrument,
            date=i1.date,
            initial_shares=1000,
        )
        a2 = asset_position_factory.create(
            portfolio=portfolio,
            portfolio__currency__key="USD",
            underlying_instrument=i1.instrument,
            date=i2.date,
            initial_shares=1000,
        )

        # Make an empty df
        request.GET = request.GET.copy()
        request.GET.update(
            {
                "historic_date": (i1.date - BDay(2)).date().strftime("%Y-%m-%d"),
                "compared_date": (i1.date - BDay(3)).date().strftime("%Y-%m-%d"),
                "bigger_than_x": 1,
            }
        )
        df_empty = AggregatedAssetPositionLiquidityPandasView(request=request)._get_dataframe()

        # Run the function with existing second_date in asset_position
        request.GET["compared_date"] = i2.date.strftime("%Y-%m-%d")

        # Run with 2 existing dates.
        request.GET["historic_date"] = i1.date.strftime("%Y-%m-%d")
        df = AggregatedAssetPositionLiquidityPandasView(request=request)._get_dataframe()
        # Coverage:
        assert df_empty.empty  # df is empty
        # Unit tests:
        s_i1 = df.set_index("id").loc[i1.instrument.id]
        assert s_i1.at["liquidity_first_date"] == round(a1.initial_shares / (i1.volume_50d * 0.33), 2)
        assert s_i1.at["liquidity_second_date"] == round(a2.initial_shares / (i2.volume_50d * 0.33), 2)



import logging
from collections import defaultdict
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from math import isclose
from typing import Any, Iterable

import numpy as np
import pandas as pd
from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import DateRangeField
from django.db import models
from django.db.models import (
    BooleanField,
    Case,
    Exists,
    F,
    OuterRef,
    Q,
    QuerySet,
    Sum,
    Value,
    When,
)
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.functional import cached_property
from pandas._libs.tslibs.offsets import BDay
from psycopg.types.range import DateRange
from skfolio.preprocessing import prices_to_returns
from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.models import WBModel
from wbcore.utils.importlib import import_from_dotted_path
from wbcore.utils.models import ActiveObjectManager, DeleteToDisableMixin
from wbfdm.contrib.metric.tasks import compute_metrics_as_task
from wbfdm.models import Instrument, InstrumentType
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbfdm.signals import investable_universe_updated

from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.indexes import Index
from wbportfolio.models.portfolio_relationship import (
    InstrumentPortfolioThroughModel,
    PortfolioInstrumentPreferredClassificationThroughModel,
)
from wbportfolio.models.products import Product
from wbportfolio.pms.analytics.portfolio import Portfolio as AnalyticPortfolio
from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from . import ProductGroup

logger = logging.getLogger("pms")


class PositionDictConverter:
    """Utility class to load a dictionary of dictionary (dictionary of instrument_id and float per date) into a iterable of assetposition"""

    def __init__(self, portfolio, prices: pd.DataFrame, infer_underlying_quote_price=False):
        self.portfolio = portfolio
        self.instruments = {}
        self.fx_rates = defaultdict(dict)
        self.prices = prices
        self.infer_underlying_quote_price = infer_underlying_quote_price

    def _get_instrument(self, instrument_id: int) -> Instrument:
        try:
            return self.instruments[instrument_id]
        except KeyError:
            instrument = Instrument.objects.get(id=instrument_id)
            self.instruments[instrument_id] = instrument
            return instrument

    def _get_fx_rate(self, val_date: date, currency) -> CurrencyFXRates:
        try:
            return self.fx_rates[currency][val_date]
        except KeyError:
            fx_rate = CurrencyFXRates.objects.get_or_create(
                currency=currency, date=val_date, defaults={"value": Decimal(0)}
            )[0]  # we create a fx rate anyway to not fail the position. The fx rate expect to be there later on
            self.fx_rates[currency][val_date] = fx_rate
            return fx_rate

    def dict_to_model(self, val_date: date, instrument_id: int, weighting: float) -> AssetPosition:
        underlying_quote = self._get_instrument(instrument_id)
        currency_fx_rate_portfolio_to_usd = self._get_fx_rate(val_date, self.portfolio.currency)
        currency_fx_rate_instrument_to_usd = self._get_fx_rate(val_date, underlying_quote.currency)
        position = AssetPosition(
            underlying_quote=underlying_quote,
            weighting=weighting,
            date=val_date,
            asset_valuation_date=val_date,
            is_estimated=True,
            portfolio=self.portfolio,
            currency=underlying_quote.currency,
            initial_price=self.prices.loc[pd.Timestamp(val_date), instrument_id],
            currency_fx_rate_portfolio_to_usd=currency_fx_rate_portfolio_to_usd,
            currency_fx_rate_instrument_to_usd=currency_fx_rate_instrument_to_usd,
            initial_currency_fx_rate=None,
            underlying_quote_price=None,
            underlying_instrument=None,
        )
        position.pre_save(
            infer_underlying_quote_price=self.infer_underlying_quote_price
        )  # inferring underlying quote price is potentially very slow for big dataset of positions, it's not very needed for model portfolio so we disable it
        return position

    def convert(self, positions: dict[date, dict[int, float]]) -> Iterable[AssetPosition]:
        for val_date, weights in positions.items():
            yield from map(lambda row: self.dict_to_model(val_date, row[0], row[1]), weights.items())


class DefaultPortfolioQueryset(QuerySet):
    def filter_invested_at_date(self, val_date: date) -> QuerySet:
        """
        Filter the queryset to get only portfolio invested at the given date
        """
        return self.filter(
            (Q(invested_timespan__startswith__lte=val_date) | Q(invested_timespan__startswith__isnull=True))
            & (Q(invested_timespan__endswith__gt=val_date) | Q(invested_timespan__endswith__isnull=True))
        )

    def to_dependency_iterator(self, val_date: date) -> Iterable["Portfolio"]:
        """
        A method to sort the given queryset to return undependable portfolio first. This is very useful if a routine needs to be applied sequentially on portfolios by order of dependence.
        """
        MAX_ITERATIONS: int = (
            5  # in order to avoid circular dependency and infinite loop, we need to stop recursion at a max depth
        )
        remaining_portfolios = set(self)

        def _iterator(p, iterator_counter=0):
            iterator_counter += 1
            parent_portfolios = remaining_portfolios & set(
                map(lambda o: o[0], p.get_parent_portfolios(val_date))
            )  # get composition parent portfolios
            dependency_relationships = PortfolioPortfolioThroughModel.objects.filter(
                portfolio=p, dependency_portfolio__in=remaining_portfolios
            )  # get dependency portfolios
            if iterator_counter >= MAX_ITERATIONS or (
                not dependency_relationships.exists() and not bool(parent_portfolios)
            ):  # if not dependency portfolio or parent portfolio that remained, then we yield
                remaining_portfolios.remove(p)
                yield p
            else:
                # otherwise, we iterate of the dependency portfolio first
                deps_portfolios = parent_portfolios.union(
                    set([r.dependency_portfolio for r in dependency_relationships])
                )
                for deps_p in deps_portfolios:
                    yield from _iterator(deps_p, iterator_counter=iterator_counter)

        while len(remaining_portfolios) > 0:
            portfolio = next(iter(remaining_portfolios))
            yield from _iterator(portfolio)


class DefaultPortfolioManager(ActiveObjectManager):
    def get_queryset(self):
        return DefaultPortfolioQueryset(self.model).filter(is_active=True)

    def filter_invested_at_date(self, val_date: date):
        return self.get_queryset().filter_invested_at_date(val_date)


class ActiveTrackedPortfolioManager(DefaultPortfolioManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(asset_exists=Exists(AssetPosition.unannotated_objects.filter(portfolio=OuterRef("pk"))))
            .filter(asset_exists=True, is_tracked=True)
        )


class PortfolioPortfolioThroughModel(models.Model):
    class Type(models.TextChoices):
        PRIMARY = "PRIMARY", "Primary"
        MODEL = "MODEL", "Model"
        CUSTODIAN = "CUSTODIAN", "Custodian"

    portfolio = models.ForeignKey("wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependency_through")
    dependency_portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependent_through"
    )
    type = models.CharField(choices=Type.choices, default=Type.PRIMARY, verbose_name="Type")

    def __str__(self):
        return f"{self.portfolio} dependant on {self.dependency_portfolio} ({self.Type[self.type].label})"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["portfolio", "type"], name="unique_primary", condition=Q(type="PRIMARY")),
            models.UniqueConstraint(fields=["portfolio", "type"], name="unique_model", condition=Q(type="MODEL")),
        ]


class Portfolio(DeleteToDisableMixin, WBModel):
    assets: models.QuerySet[AssetPosition]

    name = models.CharField(
        max_length=255,
        verbose_name="Name",
        default="",
        help_text="The Name of the Portfolio",
    )

    currency = models.ForeignKey(
        to="currency.Currency",
        related_name="portfolios",
        on_delete=models.PROTECT,
        verbose_name="Currency",
        help_text="The currency of the portfolio.",
    )
    hedged_currency = models.ForeignKey(
        to="currency.Currency",
        related_name="hedged_portfolios",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name="Hedged Currency",
        help_text="The hedged currency of the portfolio.",
    )
    depends_on = models.ManyToManyField(
        "wbportfolio.Portfolio",
        symmetrical=False,
        related_name="dependent_portfolios",
        through="wbportfolio.PortfolioPortfolioThroughModel",
        through_fields=("portfolio", "dependency_portfolio"),
        blank=True,
        verbose_name="The portfolios this portfolio depends on",
    )

    preferred_instrument_classifications = models.ManyToManyField(
        "wbfdm.Instrument",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
        related_name="preferred_portfolio_classifications",
        through="wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel",
        through_fields=("portfolio", "instrument"),
        blank=True,
        verbose_name="The Preferred classification per instrument",
    )
    instruments = models.ManyToManyField(
        "wbfdm.Instrument",
        through=InstrumentPortfolioThroughModel,
        related_name="portfolios",
        blank=True,
        verbose_name="Instruments",
        help_text="Instruments linked to this instrument",
    )
    invested_timespan = DateRangeField(
        null=True, blank=True, help_text="Define when this portfolio is considered invested"
    )

    is_manageable = models.BooleanField(
        default=False,
        help_text="True if the portfolio can be manually modified (e.g. Trade proposal be submitted or total weight recomputed)",
    )
    is_tracked = models.BooleanField(
        default=True,
        help_text="True if the internal updating mechanism (e.g., Next weights or Look-Through computation, rebalancing etc...) needs to apply to this portfolio",
    )
    only_weighting = models.BooleanField(
        default=False,
        help_text="Indicates that this portfolio is only utilizing weights and disregards shares, e.g. a model portfolio",
    )
    is_lookthrough = models.BooleanField(
        default=False,
        help_text="Indicates that this portfolio is a look-through portfolio",
    )
    is_composition = models.BooleanField(
        default=False, help_text="If true, this portfolio is a composition of other portfolio"
    )
    updated_at = models.DateTimeField(blank=True, null=True, verbose_name="Updated At")
    bank_accounts = models.ManyToManyField(
        to="directory.BankingContact",
        related_name="wbportfolio_portfolios",
        through="wbportfolio.PortfolioBankAccountThroughModel",
        blank=True,
    )

    objects = DefaultPortfolioManager()
    tracked_objects = ActiveTrackedPortfolioManager()

    @property
    def primary_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.PRIMARY
            ).dependency_portfolio

    @property
    def model_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
            ).dependency_portfolio

    @property
    def composition_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self,
                type=PortfolioPortfolioThroughModel.Type.MODEL,
                dependency_portfolio__is_composition=True,
            ).dependency_portfolio

    @property
    def imported_assets(self):
        return self.assets.filter(is_estimated=False)

    @cached_property
    def pms_instruments(self):
        instruments = [i for i in Product.objects.filter(portfolios=self)]
        instruments.extend([i for i in ProductGroup.objects.filter(portfolios=self)])
        instruments.extend([i for i in Index.objects.filter(portfolios=self)])
        return instruments

    @property
    def can_be_rebalanced(self):
        return self.is_tracked and self.is_manageable and not self.is_lookthrough

    def delete(self, **kwargs):
        super().delete(**kwargs)
        # We check if for all linked instruments, this portfolio was the last active one (if yes, we disable the instrument)
        if self.id:
            for instrument in self.instruments.iterator():
                if not instrument.portfolios.filter(is_active=True).exists():
                    instrument.delisted_date = date.today() - timedelta(days=1)
                    instrument.save()

    def _build_dto(self, val_date: date, **extra_kwargs) -> PortfolioDTO:
        "returns the dto representation of this portfolio at the specified date"
        return PortfolioDTO(
            tuple([pos._build_dto() for pos in self.assets.filter(date=val_date, **extra_kwargs)]),
        )

    def get_weights(self, val_date: date) -> dict[int, float]:
        """
        A convenience utility method to returns the portfolio weights for this portfolio as a dictionary (instrument id as key and weights as value)

        Args:
            val_date: The date at which to return the weights for this portfolio

        Returns:
            A dictionary containing the weights for this portfolio
        """
        return dict(
            map(
                lambda r: (r[0], float(r[1])),
                self.assets.filter(date=val_date)
                .values("underlying_quote")
                .annotate(sum_weight=Sum("weighting"))
                .values_list("underlying_quote", "sum_weight"),
            )
        )

    def get_analytic_portfolio(self, val_date: date, **kwargs) -> tuple[AnalyticPortfolio, pd.DataFrame]:
        """
        Return the analytic portfolio associated with this portfolio at the given date

        the analytic portfolio inherit from SKFolio Portfolio and can be used to access all this library methods
        Args:
            val_date: the date to calculate the portfolio for

        Returns:
            The instantiated analytic portfolio
        """
        weights = self.get_weights(val_date)
        instrument_ids = weights.keys()
        return_date = (val_date + BDay(1)).date()
        returns, prices = self.get_returns(instrument_ids, (val_date - BDay(2)).date(), return_date, **kwargs)
        returns = returns.fillna(0)  # not sure this is what we want
        return (
            AnalyticPortfolio(
                X=returns,
                weights=weights,
            ),
            prices,
        )

    def is_invested_at_date(self, val_date: date) -> bool:
        return (
            self.invested_timespan
            and self.invested_timespan.upper > val_date
            and self.invested_timespan.lower <= val_date
        )

    def __str__(self):
        return f"{self.id:06} ({self.name})"

    class Meta:
        verbose_name = "Portfolio"
        verbose_name_plural = "Portfolios"

        notification_types = [
            create_notification_type(
                "wbportfolio.portfolio.check_custodian_portfolio",
                "Check Custodian Portfolio",
                "Sends a notification when a portfolio does not match with its custodian portfolio",
                True,
                True,
                True,
            ),
        ]

    def is_active_at_date(self, val_date: date) -> bool:
        """
        Return if the base instrument has a total aum greater than 0
        :val_date: the date at which we need to evaluate if the portfolio is considered active
        """
        active_portfolio = self.is_active or self.deletion_datetime.date() > val_date
        if self.instruments.exists():
            return active_portfolio and any(
                [instrument.is_active_at_date(val_date) for instrument in self.instruments.all()]
            )
        return active_portfolio

    def get_total_asset_value(self, val_date: date) -> Decimal:
        """
        Return the total asset under management of the portfolio at the specified valuation date
        Args:
            val_date: The date at which aum needs to be computed
        Returns:
            The total AUM (0 if there is no position)
        """
        return self.assets.filter(date=val_date).aggregate(s=Sum("total_value_fx_portfolio"))["s"] or Decimal(0.0)

    def get_total_asset_under_management(self, val_date):
        from wbportfolio.models.transactions.trades import Trade

        trades = Trade.valid_customer_trade_objects.filter(portfolio=self, transaction_date__lte=val_date)

        total_aum = Decimal(0)
        for underlying_instrument_id, sum_shares in (
            trades.values("underlying_instrument")
            .annotate(
                sum_shares=Sum("shares"),
            )
            .values_list("underlying_instrument", "sum_shares")
        ):
            with suppress(Instrument.DoesNotExist, InstrumentPrice.DoesNotExist):
                instrument = Instrument.objects.get(id=underlying_instrument_id)
                last_price = instrument.valuations.filter(date__lte=val_date).latest("date").net_value
                fx_rate = instrument.currency.convert(val_date, self.currency)
                total_aum += last_price * sum_shares * fx_rate
        return total_aum

    def _get_assets(self, with_estimated=True, with_cash=True):
        qs = self.assets
        if not with_estimated:
            qs = qs.filter(is_estimated=False)
        if not with_cash:
            qs = qs.exclude(underlying_instrument__is_cash=True)
        return qs

    def get_earliest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__gte=val_date)
        if qs.exists():
            return qs.earliest("date").date
        return None

    def get_latest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__lte=val_date)

        if qs.exists():
            return qs.latest("date").date
        return None

    # Asset Position Utility Functions
    def get_holding(self, val_date, exclude_cash=True, exclude_index=True):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date, weighting__gt=0)
        if exclude_index:
            qs = qs.exclude(underlying_instrument__instrument_type=InstrumentType.INDEX)
        return (
            qs.values("underlying_instrument__name")
            .annotate(total_value_fx_portfolio=Sum("total_value_fx_portfolio"), weighting=Sum("weighting"))
            .order_by("-total_value_fx_portfolio")
        )

    def _get_groupedby_df(
        self,
        group_by,
        val_date: date,
        exclude_cash: bool | None = False,
        exclude_index: bool | None = False,
        extra_filter_parameters: dict[str, Any] = None,
        **groupby_kwargs,
    ):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date)
        if exclude_index:
            # We exclude only index that are not considered as cash. Setting exclude_cash to true convers this case.
            qs = qs.exclude(
                Q(underlying_instrument__instrument_type=InstrumentType.INDEX)
                & Q(underlying_instrument__is_cash=False)
            )
        if extra_filter_parameters:
            qs = qs.filter(**extra_filter_parameters)
        qs = group_by(qs, **groupby_kwargs).annotate(sum_weighting=Sum(F("weighting"))).order_by("-sum_weighting")
        df = pd.DataFrame(
            qs.values_list("aggregated_title", "sum_weighting"), columns=["aggregated_title", "weighting"]
        )
        if not df.empty:
            df.weighting = df.weighting.astype("float")
            df.weighting = df.weighting / df.weighting.sum()
            df = df.sort_values(by=["weighting"])
        return df.where(pd.notnull(df), None)

    def get_geographical_breakdown(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.country_group_by, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_currency_exposure(self, val_date, **kwargs):
        df = self._get_groupedby_df(AssetPosition.currency_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_market_cap_distribution(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.marketcap_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_instrument__instrument_type": InstrumentType.EQUITY},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_liquidity(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.liquidity_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_instrument__instrument_type": InstrumentType.EQUITY},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_industry_exposure(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.group_by_primary, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_asset_allocation(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(AssetPosition.cash_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_adjusted_child_positions(self, val_date):
        if (
            child_positions := self.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        ).count() == 1:
            if portfolio := child_positions.first().underlying_instrument.primary_portfolio:
                child_positions = portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        for position in child_positions:
            if child_portfolio := position.underlying_instrument.primary_portfolio:
                index_positions = child_portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(
                    date=val_date
                )

                for index_position in index_positions.all():
                    weighting = index_position.weighting * position.weighting
                    if weighting != 0:
                        yield {
                            "underlying_instrument_id": index_position.underlying_instrument.id,
                            "weighting": weighting,
                        }

    def get_longshort_distribution(self, val_date):
        df = pd.DataFrame(self.get_adjusted_child_positions(val_date))

        if not df.empty:
            df["is_cash"] = df.underlying_instrument_id.apply(lambda x: Instrument.objects.get(id=x).is_cash)
            df = df[~df["is_cash"]]
            df = (
                df[["underlying_instrument_id", "weighting"]].groupby("underlying_instrument_id").sum().astype("float")
            )
            df.weighting = df.weighting / df.weighting.sum()
            short_weight = df[df.weighting < 0].weighting.abs().sum()
            long_weight = df[df.weighting > 0].weighting.sum()
            total_weight = long_weight + short_weight
            return pd.DataFrame(
                [
                    {"title": "Long", "weighting": long_weight / total_weight},
                    {"title": "Short", "weighting": short_weight / total_weight},
                ]
            )
        return df

    def get_portfolio_contribution_df(self, start, end, with_cash=True, hedged_currency=None, only_equity=False):
        qs = self._get_assets(with_cash=with_cash).filter(date__gte=start, date__lte=end)
        if only_equity:
            qs = qs.filter(underlying_instrument__instrument_type=InstrumentType.EQUITY)
        return Portfolio.get_contribution_df(qs, hedged_currency=hedged_currency)

    def check_related_portfolio_at_date(self, val_date: date, related_portfolio: "Portfolio"):
        assets = AssetPosition.objects.filter(
            date=val_date, underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ).values("underlying_instrument", "shares")
        assets1 = assets.filter(portfolio=self)
        assets2 = assets.filter(portfolio=related_portfolio)
        return assets1.difference(assets2)

    def get_child_portfolios(self, val_date: date) -> set["Portfolio"]:
        child_portfolios = set()
        instrument_rel = InstrumentPortfolioThroughModel.objects.filter(portfolio=self)
        if instrument_rel.exists():
            for parent_portfolio in Portfolio.objects.filter(
                id__in=AssetPosition.unannotated_objects.filter(
                    date=val_date, underlying_quote__in=instrument_rel.values("instrument")
                ).values("portfolio")
            ):
                child_portfolios.add(parent_portfolio)
        return child_portfolios

    def get_parent_portfolios(self, val_date: date) -> set["Portfolio"]:
        for asset in self.assets.filter(date=val_date, underlying_instrument__portfolios__isnull=False).distinct(
            "underlying_instrument"
        ):
            if portfolio := asset.underlying_instrument.portfolio:
                yield portfolio, asset.weighting

    def change_at_date(
        self,
        val_date: date,
        recompute_weighting: bool = False,
        force_recompute_weighting: bool = False,
        compute_metrics: bool = False,
        evaluate_rebalancer: bool = True,
    ):
        logger.info(f"change at date for {self} at {val_date}")

        if recompute_weighting:
            # We normalize weight across the portfolio for a given date
            qs = self.assets.filter(date=val_date).filter(
                Q(total_value_fx_portfolio__isnull=False) | Q(weighting__isnull=False)
            )
            if (self.is_lookthrough or self.is_manageable or force_recompute_weighting) and qs.exists():
                total_weighting = qs.aggregate(s=Sum("weighting"))["s"]
                # We check if this actually necessary
                # (i.e. if the weight is already summed to 100%, it is already normalized)
                if (
                    not total_weighting
                    or not isclose(total_weighting, Decimal(1.0), abs_tol=0.001)
                    or force_recompute_weighting
                ):
                    total_value = qs.aggregate(s=Sum("total_value_fx_portfolio"))["s"]
                    # TODO we change this because postgres doesn't support join statement in update (and total_value_fx_portfolio is a joined annoted field)
                    for asset in qs:
                        if total_value:
                            asset.weighting = asset._total_value_fx_portfolio / total_value
                        elif total_weighting:
                            asset.weighting = asset.weighting / total_weighting
                        asset.save()

        # We check if there is an instrument attached to the portfolio with calculated NAV and price computation method
        self.estimate_net_asset_values((val_date + BDay(1)).date())  # updating weighting in t0 influence nav in t+1
        if evaluate_rebalancer:
            self.evaluate_rebalancing(val_date)

        self.updated_at = timezone.now()
        self.save()

        if compute_metrics:
            compute_metrics_as_task.delay(
                val_date, basket_id=self.id, basket_content_type_id=ContentType.objects.get_for_model(Portfolio).id
            )
        self.handle_controlling_portfolio_change_at_date(val_date)

    def handle_controlling_portfolio_change_at_date(self, val_date: date):
        for rel in PortfolioPortfolioThroughModel.objects.filter(
            dependency_portfolio=self, type=PortfolioPortfolioThroughModel.Type.PRIMARY, portfolio__is_lookthrough=True
        ):
            portfolio_total_asset_value = (
                self.get_total_asset_under_management(val_date) if not self.only_weighting else None
            )
            rel.portfolio.compute_lookthrough(val_date, portfolio_total_asset_value=portfolio_total_asset_value)
        for rel in PortfolioPortfolioThroughModel.objects.filter(
            dependency_portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
        ):
            rel.portfolio.evaluate_rebalancing(val_date)
        for dependent_portfolio in self.get_child_portfolios(val_date):
            dependent_portfolio.handle_controlling_portfolio_change_at_date(val_date)

    def evaluate_rebalancing(self, val_date: date):
        if hasattr(self, "automatic_rebalancer"):
            # if the portfolio has an automatic rebalancer and the next business day is suitable with the rebalancer, we create a trade proposal automatically
            next_business_date = (val_date + BDay(1)).date()
            if self.automatic_rebalancer.is_valid(val_date):  # we evaluate the rebalancer in t0 and t+1
                logger.info(f"Evaluate Rebalancing for {self} at {val_date}")
                self.automatic_rebalancer.evaluate_rebalancing(val_date)
            if self.automatic_rebalancer.is_valid(next_business_date):
                logger.info(f"Evaluate Rebalancing for {self} at {next_business_date}")
                self.automatic_rebalancer.evaluate_rebalancing(next_business_date)

    def estimate_net_asset_values(self, val_date: date):
        for instrument in self.pms_instruments:
            if instrument.is_active_at_date(val_date) and (
                net_asset_value_computation_method_path := instrument.net_asset_value_computation_method_path
            ):
                logger.info(f"Estimate NAV of {val_date:%Y-%m-%d} for instrument {instrument}")
                net_asset_value_computation_method = import_from_dotted_path(net_asset_value_computation_method_path)
                estimated_net_asset_value = net_asset_value_computation_method(val_date, instrument)
                if estimated_net_asset_value is not None:
                    InstrumentPrice.objects.update_or_create(
                        instrument=instrument,
                        date=val_date,
                        calculated=True,
                        defaults={
                            "gross_value": estimated_net_asset_value,
                            "net_value": estimated_net_asset_value,
                        },
                    )
                    if (
                        val_date == instrument.prices.latest("date").date
                    ):  # if price date is the latest instrument price date, we recompute the last valuation data
                        instrument.update_last_valuation_date()
                instrument.update_last_valuation_date()

    def batch_portfolio(self, start_date: date, end_date: date):
        """
        Create the cumulative portfolios between the two given dates and stop at the first rebalancing (if any)

        Returns: The trade proposal generated by the rebalancing, if any (otherwise None)
        """
        rebalancer = getattr(self, "automatic_rebalancer", None)

        weights = self.get_weights(start_date)  # initial weights
        positions = dict()
        next_trade_proposal = None
        rebalancing_date = None
        logger.info(f"compute next weights in batch for {self} from {start_date:%Y-%m-%d} to {end_date:%Y-%m-%d}")
        returns, prices = self.get_returns(weights.keys(), (start_date - BDay(3)).date(), end_date, ffill_returns=True)
        for to_date_ts in pd.date_range(start_date + timedelta(days=1), end_date, freq="B"):
            to_date = to_date_ts.date()
            logger.info(f"Processing {to_date:%Y-%m-%d}")
            if rebalancer and rebalancer.is_valid(to_date):
                rebalancing_date = to_date
                break
            last_returns = returns.loc[[to_date_ts], :]
            analytic_portfolio = AnalyticPortfolio(weights=weights, X=last_returns)
            positions[to_date] = analytic_portfolio.get_next_weights()

        positions_generator = PositionDictConverter(self, prices)
        positions = list(positions_generator.convert(positions))
        self.bulk_create_positions(positions, delete_leftovers=True, compute_metrics=False, evaluate_rebalancer=False)
        if rebalancing_date:
            next_trade_proposal = rebalancer.evaluate_rebalancing(rebalancing_date)

        return next_trade_proposal

    def propagate_or_update_assets(self, from_date: date, to_date: date):
        """
        Create a new portfolio at `to_date` based on the portfolio in `from_date`.

        Args:
            from_date: The date to propagate the portfolio from
            to_date:  The date to create the new portfolio at

        """
        from_is_active = self.is_active_at_date(from_date)
        to_is_active = self.is_active_at_date(to_date)

        def _parse_position(asset: AssetPosition) -> AssetPosition:
            # function to handle the position modification after instantiation
            if from_is_active and not to_is_active:
                asset.weighting = Decimal(0.0)
            asset.initial_shares = AssetPosition.unannotated_objects.filter(
                date=from_date, underlying_quote=asset.underlying_quote, portfolio=self
            ).aggregate(sum_shares=Sum("initial_shares"))["sum_shares"]
            return asset

        # we don't propagate on already imported portfolio by default
        is_target_portfolio_imported = self.assets.filter(date=to_date, is_estimated=False).exists()
        if (
            self.is_tracked and not self.is_lookthrough and not is_target_portfolio_imported and from_is_active
        ):  # we cannot propagate a new portfolio for untracked, or look-through or already imported or inactive portfolios
            logger.info(f"computing next weight for {self} from {from_date:%Y-%m-%d} to {to_date:%Y-%m-%d}")
            analytic_portfolio, to_prices = self.get_analytic_portfolio(from_date)
            if not to_prices.empty:
                weights = analytic_portfolio.get_next_weights()
                positions_generator = PositionDictConverter(
                    self, to_prices, infer_underlying_quote_price=True
                ).convert({to_date: weights})
                positions = list(map(lambda a: _parse_position(a), positions_generator))
                self.bulk_create_positions(positions, delete_leftovers=True, compute_metrics=True)

    def get_lookthrough_positions(
        self,
        sync_date: date,
        portfolio_total_asset_value: Decimal | None = None,
        with_intermediary_position: bool = False,
    ):
        """Recursively calculates the look-through position for a portfolio

        Arguments:
            sync_date {datetime.date} -- The date on which the assets will be computed
            portfolio_total_value: {Decimal} -- The total value of the portfolio (needed to compute initial shares)
        """

        def _crawl_portfolio(
            parent_portfolio,
            adjusted_weighting,
            adjusted_currency_fx_rate,
            adjusted_is_estimated,
            portfolio_created=None,
            path=None,
        ):
            if not path:
                path = []
            path.append(parent_portfolio)
            for position in parent_portfolio.assets.filter(date=sync_date):
                position.id = None
                position.weighting = adjusted_weighting * position.weighting
                position.initial_currency_fx_rate = adjusted_currency_fx_rate * position.currency_fx_rate
                position.is_estimated = (adjusted_is_estimated or position.is_estimated) and not (
                    position.weighting == 1.0
                )
                position.portfolio_created = portfolio_created
                setattr(position, "path", path)
                position.initial_shares = None
                if portfolio_total_asset_value:
                    position.initial_shares = (position.weighting * portfolio_total_asset_value) / (
                        position.price * position.currency_fx_rate
                    )
                if child_portfolio := position.underlying_quote.primary_portfolio:
                    if with_intermediary_position:
                        yield position
                    yield from _crawl_portfolio(
                        child_portfolio,
                        position.weighting,
                        position.currency_fx_rate,
                        position.is_estimated,
                        portfolio_created=child_portfolio,
                        path=path.copy(),
                    )
                elif position.weighting:  # we do not yield position with weight 0 because of issue with certain multi-thematic portfolios which contain duplicates
                    yield position

        yield from _crawl_portfolio(self, Decimal(1.0), Decimal(1.0), False)

    def get_positions(self, val_date: date, **kwargs) -> Iterable[AssetPosition]:
        if self.is_composition:
            assets = list(self.get_lookthrough_positions(val_date, **kwargs))
        else:
            assets = self.assets.filter(date=val_date)
        return assets

    def compute_lookthrough(self, sync_date: date, portfolio_total_asset_value: Decimal | None = None):
        if not self.primary_portfolio or not self.is_lookthrough:
            raise ValueError(
                "Lookthrough position can only be computed on lookthrough portfolio with a primary portfolio"
            )
        logger.info(f"Compute Look-Through for {self} at {sync_date}")
        positions = self.primary_portfolio.get_lookthrough_positions(sync_date, portfolio_total_asset_value)
        self.bulk_create_positions(list(positions), delete_leftovers=True, compute_metrics=True)

    def batch_recompute_lookthrough(self, from_date: date, to_date: date):
        if not self.primary_portfolio or not self.is_lookthrough:
            raise ValueError(
                "Lookthrough position can only be computed on lookthrough portfolio with a primary portfolio"
            )
        positions = []
        for val_date in pd.date_range(from_date, to_date, freq="B").date:
            portfolio_total_asset_value = (
                self.get_total_asset_under_management(val_date) if not self.only_weighting else None
            )
            positions.extend(self.primary_portfolio.get_lookthrough_positions(val_date, portfolio_total_asset_value))
        self.bulk_create_positions(list(positions), delete_leftovers=True, compute_metrics=False)

    def update_preferred_classification_per_instrument(self):
        # Function to automatically assign Preferred instrument based on the assets' underlying instruments of the
        # attached wbportfolio
        instruments = filter(
            None,
            map(
                lambda x: Instrument.objects.get(id=x["underlying_instrument"]).get_classifable_ancestor(
                    include_self=True
                ),
                self.assets.values("underlying_instrument").distinct("underlying_instrument"),
            ),
        )
        leftovers_instruments = list(
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(portfolio=self).values_list(
                "instrument", flat=True
            )
        )
        for instrument in instruments:
            other_classifications = instrument.classifications.filter(group__is_primary=False)
            default_classification = None
            if other_classifications.count() == 1:
                default_classification = other_classifications.first()
            if not PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument
            ).exists():
                PortfolioInstrumentPreferredClassificationThroughModel.objects.create(
                    portfolio=self,
                    instrument=instrument,
                    classification=default_classification,
                    classification_group=default_classification.group if default_classification else None,
                )
            if instrument.id in leftovers_instruments:
                leftovers_instruments.remove(instrument.id)

        for instrument_id in leftovers_instruments:
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument_id
            ).delete()

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfolio"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:portfoliorepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    def bulk_create_positions(self, positions: list[AssetPosition], delete_leftovers: bool = False, **kwargs):
        if positions:
            update_dates = set()
            for position in positions:
                position.portfolio = self
                update_dates.add(position.date)
            # self.assets.filter(date__in=update_dates, is_estimated=True).delete()
            leftover_positions_ids = list(
                self.assets.filter(date__in=update_dates).values_list("id", flat=True)
            )  # we need to get the ids otherwise the queryset is reevaluated later
            logger.info(f"bulk saving {len(positions)} positions ({len(leftover_positions_ids)} leftovers) ...")
            objs = AssetPosition.unannotated_objects.bulk_create(
                positions,
                update_fields=[
                    "weighting",
                    "initial_price",
                    "initial_currency_fx_rate",
                    "initial_shares",
                    "currency_fx_rate_instrument_to_usd",
                    "currency_fx_rate_portfolio_to_usd",
                    "underlying_quote_price",
                    "is_estimated",
                    "portfolio",
                    "portfolio_created",
                    "underlying_instrument",
                ],
                unique_fields=["portfolio", "date", "underlying_quote", "portfolio_created"],
                update_conflicts=True,
                batch_size=10000,
            )
            if delete_leftovers:
                objs_ids = list(map(lambda x: x.id, objs))
                leftover_positions_ids = list(filter(lambda i: i not in objs_ids, leftover_positions_ids))
                logger.info(f"deleting {len(leftover_positions_ids)} leftover positions..")
                AssetPosition.objects.filter(id__in=leftover_positions_ids).delete()
            for update_date in sorted(update_dates):
                self.change_at_date(update_date, **kwargs)

    @classmethod
    def _get_or_create_portfolio(cls, instrument_handler, portfolio_data):
        if isinstance(portfolio_data, int):
            return Portfolio.objects.get(id=portfolio_data)
        instrument = portfolio_data
        if isinstance(portfolio_data, dict):
            instrument = instrument_handler.process_object(instrument, only_security=False, read_only=True)[0]
        return instrument.primary_portfolio

    def check_share_diff(self, val_date: date) -> bool:
        return self.assets.filter(Q(date=val_date) & ~Q(initial_shares=F("initial_shares_at_custodian"))).exists()

    def get_returns(
        self,
        instruments: Iterable,
        from_date: date,
        to_date: date,
        ffill_returns: bool = True,
        convert_to_portfolio_currency: bool = True,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Utility methods to get instrument returns for a given date range

        Args:
            instruments: instrument to get the returns from
            from_date: date range lower bound
            to_date: date range upper bound

        Returns:
            Return a tuple of the returns and the last prices series for conveniance
        """
        prices = InstrumentPrice.objects.filter(
            instrument__in=instruments, date__gte=from_date, date__lte=to_date
        ).annotate(
            fx_rate=CurrencyFXRates.get_fx_rates_subquery_for_two_currencies(
                "date", "instrument__currency", self.currency
            ),
            price_fx_portfolio=F("net_value") * F("fx_rate") if convert_to_portfolio_currency else F("net_value"),
        )
        prices_df = (
            pd.DataFrame(
                prices.filter_only_valid_prices().values_list("instrument", "price_fx_portfolio", "date"),
                columns=["instrument", "price_fx_portfolio", "date"],
            )
            .pivot_table(index="date", values="price_fx_portfolio", columns="instrument")
            .astype(float)
            .sort_index()
        )
        ts = pd.bdate_range(from_date, to_date, freq="B")
        prices_df = prices_df.reindex(ts)
        if ffill_returns:
            prices_df = prices_df.ffill()
        returns = prices_to_returns(prices_df, drop_inceptions_nan=False, fill_nan=ffill_returns)
        return returns.replace([np.inf, -np.inf, np.nan], 0), prices_df

    @classmethod
    def get_contribution_df(
        cls,
        qs,
        need_normalize=False,
        groupby_label_id="underlying_instrument",
        groubpy_label_title="underlying_instrument__name_repr",
        currency_fx_rate_label="currency_fx_rate",
        hedged_currency=None,
    ):
        # qs = AssetPosition.annotate_underlying_instrument(qs)
        weight_label = "weighting" if not need_normalize else "total_value_fx_portfolio"
        qs = qs.annotate(
            is_hedged=Case(
                When(
                    underlying_instrument__currency__isnull=False,
                    underlying_instrument__currency=hedged_currency,
                    then=Value(True),
                ),
                default=Value(False),
                output_field=BooleanField(),
            ),
            coalesce_currency_fx_rate=Case(
                When(is_hedged=True, then=Value(Decimal(1.0))),
                default=F(currency_fx_rate_label),
                output_field=models.BooleanField(),
            ),
        ).select_related("underlying_instrument")
        df = pd.DataFrame(
            qs.values_list(
                "date",
                "price",
                "coalesce_currency_fx_rate",
                groupby_label_id,
                groubpy_label_title,
                weight_label,
            ),
            columns=[
                "date",
                "price",
                "coalesce_currency_fx_rate",
                groupby_label_id,
                groubpy_label_title,
                weight_label,
            ],
        )
        if not df.empty:
            df = df[df[weight_label] != 0]
            df.date = pd.to_datetime(df.date)
            df["price_fx_portfolio"] = df.price * df.coalesce_currency_fx_rate

            df[["price", "price_fx_portfolio", weight_label, "coalesce_currency_fx_rate"]] = df[
                ["price", "price_fx_portfolio", weight_label, "coalesce_currency_fx_rate"]
            ].astype("float")

            df[groupby_label_id] = df[groupby_label_id].fillna(0)
            df[groubpy_label_title] = df[groubpy_label_title].fillna("N/A")
            df_static = df[[groupby_label_id, groubpy_label_title]].groupby(groupby_label_id, dropna=False).first()

            df = (
                df[
                    [
                        groupby_label_id,
                        "date",
                        "price",
                        "price_fx_portfolio",
                        weight_label,
                        "coalesce_currency_fx_rate",
                    ]
                ]
                .groupby(["date", groupby_label_id], dropna=False)
                .agg(
                    {
                        "price": "mean",
                        "price_fx_portfolio": "mean",
                        weight_label: "sum",
                        "coalesce_currency_fx_rate": "mean",
                    }
                )
                .reset_index()
                .set_index("date")
                .sort_index()
            )
            df[weight_label] = df[weight_label].fillna(0)
            value = df.pivot_table(
                index="date",
                columns=[groupby_label_id],
                values=weight_label,
                fill_value=0,
                aggfunc="sum",
            )
            weights_ = value
            if need_normalize:
                total_value_price = df[weight_label].groupby("date", dropna=False).sum()
                weights_ = value.divide(total_value_price, axis=0)
            prices_usd = (
                df.pivot_table(
                    index="date",
                    columns=[groupby_label_id],
                    values="price_fx_portfolio",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            rates_fx = (
                df.pivot_table(
                    index="date",
                    columns=[groupby_label_id],
                    values="coalesce_currency_fx_rate",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            prices_usd = prices_usd.ffill()
            performance_prices = prices_usd / prices_usd.shift(1, axis=0) - 1
            contributions_prices = performance_prices.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_prices = (1 + contributions_prices.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_prices = contributions_prices.multiply(total_contrib_prices, axis=0).sum(skipna=False)
            monthly_perf_prices = (1 + performance_prices).dropna(how="all").product(axis=0, skipna=False) - 1

            rates_fx = rates_fx.ffill()
            performance_rates_fx = rates_fx / rates_fx.shift(1, axis=0) - 1
            contributions_rates_fx = performance_rates_fx.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_rates_fx = (1 + contributions_rates_fx.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_rates_fx = contributions_rates_fx.multiply(total_contrib_rates_fx, axis=0).sum(skipna=False)
            monthly_perf_rates_fx = (1 + performance_rates_fx).dropna(how="all").product(axis=0, skipna=False) - 1

            res = pd.concat(
                [
                    df_static,
                    monthly_perf_prices,
                    monthly_perf_rates_fx,
                    contributions_prices,
                    contributions_rates_fx,
                    weights_.iloc[0, :],
                    weights_.iloc[-1, :],
                    value.iloc[0, :],
                    value.iloc[-1, :],
                ],
                axis=1,
            ).reset_index()
            res.columns = [
                groupby_label_id,
                groubpy_label_title,
                "performance_total",
                "performance_forex",
                "contribution_total",
                "contribution_forex",
                "allocation_start",
                "allocation_end",
                "total_value_start",
                "total_value_end",
            ]

            return res.replace([np.inf, -np.inf, np.nan], 0)
        return pd.DataFrame()

    def get_or_create_index(self):
        index = Index.objects.create(name=self.name, currency=self.currency)
        index.portfolios.all().delete()
        InstrumentPortfolioThroughModel.objects.update_or_create(instrument=index, defaults={"portfolio": self})

    @classmethod
    def create_model_portfolio(cls, name: str, currency: Currency, with_index: bool = True):
        portfolio = cls.objects.create(
            is_manageable=True,
            name=name,
            currency=currency,
        )
        if with_index:
            portfolio.get_or_create_index()
        return portfolio


def default_estimate_net_value(val_date: date, instrument: Instrument) -> float | None:
    portfolio = instrument.portfolio
    if (previous_date := portfolio.get_latest_asset_position_date(val_date - BDay(1), with_estimated=True)) and (
        last_price := instrument.get_latest_price(previous_date)
    ):
        with suppress(IndexError):  # we silent any indexerror introduced by no returns for the past days
            analytic_portfolio, _ = portfolio.get_analytic_portfolio(previous_date)
            return analytic_portfolio.get_estimate_net_value(float(last_price.net_value))


@receiver(post_save, sender="wbportfolio.Product")
@receiver(post_save, sender="wbportfolio.ProductGroup")
def post_product_creation(sender, instance, created, raw, **kwargs):
    if not raw and (created or not InstrumentPortfolioThroughModel.objects.filter(instrument=instance).exists()):
        portfolio = Portfolio.objects.create(
            name=f"Portfolio: {instance.name}",
            currency=instance.currency,
            invested_timespan=DateRange(instance.inception_date if instance.inception_date else date.min, date.max),
        )
        InstrumentPortfolioThroughModel.objects.get_or_create(instrument=instance, defaults={"portfolio": portfolio})


@receiver(post_save, sender="wbportfolio.PortfolioPortfolioThroughModel")
def post_portfolio_relationship_creation(sender, instance, created, raw, **kwargs):
    if (
        not raw
        and created
        and instance.portfolio.is_lookthrough
        and instance.type == PortfolioPortfolioThroughModel.Type.PRIMARY
    ):
        with suppress(AssetPosition.DoesNotExist):
            earliest_primary_position_date = instance.dependency_portfolio.assets.earliest("date").date
            batch_recompute_lookthrough_as_task.delay(
                instance.portfolio.id, earliest_primary_position_date, date.today()
            )


@shared_task(queue="portfolio")
def trigger_portfolio_change_as_task(portfolio_id, val_date, **kwargs):
    portfolio = Portfolio.all_objects.get(id=portfolio_id)
    portfolio.change_at_date(val_date, **kwargs)


@shared_task(queue="portfolio")
def batch_recompute_lookthrough_as_task(portfolio_id: int, start: date, end: date):
    portfolio = Portfolio.objects.get(id=portfolio_id)
    portfolio.batch_recompute_lookthrough(start, end)


@receiver(investable_universe_updated, sender="wbfdm.Instrument")
def update_portfolio_after_investable_universe(*args, end_date: date | None = None, **kwargs):
    if not end_date:
        end_date = date.today()
    end_date = (end_date + timedelta(days=1) - BDay(1)).date()  # shift in case of business day
    from_date = (end_date - BDay(1)).date()
    for portfolio in Portfolio.tracked_objects.filter(is_lookthrough=False).to_dependency_iterator(from_date):
        try:
            portfolio.propagate_or_update_assets(from_date, end_date)
        except Exception as e:
            logger.error(f"Exception while propagating portfolio assets {portfolio}: {e}")
        portfolio.estimate_net_asset_values(end_date)



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

PRODUCTPERFORMANCELIST_MENUITEM = MenuItem(
    label="Performances",
    endpoint="wbportfolio:productperformance-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbportfolio.view_product"]
    ),
)

PRODUCTPERFORMANCECOMPARISONLIST_MENUITEM = MenuItem(
    label="Performances Summary",
    endpoint="wbportfolio:productperformancecomparison-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbportfolio.view_product"]
    ),
)
PRODUCTPERFORMANCENNMLIST_MENUITEM = MenuItem(
    label="Net new money",
    endpoint="wbportfolio:productperformancennmlist-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbportfolio.view_product"]
    ),
)



import pandas as pd

from wbportfolio.models import Trade

from .utils import get_portfolio_id

FIELD_MAP = {
    "Transaction_Date": "transaction_date",
    "Trade_Date": "book_date",
    "Value_Date": "value_date",
    "Price": "price",
    "Quantity": "shares",
    "Currency": "currency__key",
    "Isin": "underlying_instrument__isin",
    "Asset_Text": "underlying_instrument__name",
    "Portfolio_Identifier": "portfolio",
    "Booking_comment": "comment",
}


def parse(import_source):
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=";")
    df = df.loc[(df["Order_Type"] == "stex_buy_secondary") & (df["Bookkind"] == "trade_asset"), :]
    data = []
    if not df.empty:
        df = df.rename(columns=FIELD_MAP)
        df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
        df.shares = -df.shares
        df["transaction_date"] = pd.to_datetime(df["transaction_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["book_date"] = pd.to_datetime(df["book_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["value_date"] = pd.to_datetime(df["value_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["transaction_subtype"] = df.shares.apply(lambda x: Trade.Type.SELL if x < 0 else Trade.Type.BUY)
        df["portfolio"] = df["portfolio"].apply(lambda x: get_portfolio_id(x))
        df = df.dropna(subset=["underlying_instrument__isin"])

        df["underlying_instrument__currency__key"] = df["currency__key"]
        df["bank"] = "N/A"
        data = df.to_dict("records")
    return {"data": data}



import rstr
from faker import Faker
from import_export import fields
from import_export.widgets import ForeignKeyWidget
from wbcore.contrib.io.resources import FilterModelResource
from wbfdm.models import Instrument

from wbportfolio.models import Trade, TradeProposal

fake = Faker()


class TradeProposalTradeResource(FilterModelResource):
    """
    Trade Resource class to use to import trade from the trade proposal
    """

    def __init__(self, **kwargs):
        self.trade_proposal = TradeProposal.objects.get(pk=kwargs["trade_proposal_id"])
        super().__init__(**kwargs)

    def before_import(self, dataset, **kwargs):
        Trade.objects.filter(trade_proposal=self.trade_proposal).delete()

    def get_or_init_instance(self, instance_loader, row):
        try:
            return Trade.objects.get(
                trade_proposal=self.trade_proposal, underlying_instrument=row["underlying_instrument"]
            )
        except Trade.DoesNotExist:
            return Trade(
                trade_proposal=self.trade_proposal,
                underlying_instrument=row["underlying_instrument"],
                transaction_subtype=Trade.Type.BUY if row["weighting"] > 0 else Trade.Type.SELL,
                currency=row["underlying_instrument"].currency,
                transaction_date=self.trade_proposal.trade_date,
                portfolio=self.trade_proposal.portfolio,
                weighting=row["weighting"],
                status=Trade.Status.DRAFT,
            )

    DUMMY_FIELD_MAP = {
        "underlying_instrument": lambda: rstr.xeger("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})"),
        "weighting": 1.0,
        "shares": 1000.2536,
        "comment": lambda: fake.sentence(),
        "order": 1,
    }
    underlying_instrument = fields.Field(
        column_name="underlying_instrument",
        attribute="underlying_instrument",
        widget=ForeignKeyWidget(Instrument, field="isin"),
    )

    class Meta:
        import_id_fields = ("underlying_instrument",)
        fields = (
            "id",
            "underlying_instrument",
            "weighting",
            "shares",
            "comment",
            "order",
        )
        export_order = fields
        model = Trade



from decimal import Decimal

from django.core.exceptions import ObjectDoesNotExist

from wbportfolio.pms.typing import Portfolio, Position
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Composite Rebalancing")
class CompositeRebalancing(AbstractRebalancingModel):
    @property
    def base_assets(self) -> dict[int, Decimal]:
        try:
            latest_trade_proposal = self.portfolio.trade_proposals.filter(
                status="APPROVED", trade_date__lte=self.trade_date
            ).latest("trade_date")
            return latest_trade_proposal.base_assets
        except ObjectDoesNotExist:
            return dict()

    def is_valid(self) -> bool:
        return len(self.base_assets.keys()) > 0

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        for underlying_instrument, weighting in self.base_assets.items():
            positions.append(
                Position(underlying_instrument=underlying_instrument, weighting=weighting, date=self.trade_date)
            )
        return Portfolio(positions=tuple(positions))



import json

BASE_MAPPING = {
    "isin": "instrument__isin",
    "certificatePosition": "outstanding_shares",
    "currency": "instrument__currency__key",
    "aum": "market_capitalization",
    "rpl": "net_value",
}


def parse(import_source):
    content = json.load(import_source.file)
    data = {}
    if (amc := content.get("amc", None)) and (val_date := content.get("asOfDate", None)):
        data = {"date": val_date}
        for k, v in amc.items():
            if k in BASE_MAPPING:
                data[BASE_MAPPING[k]] = v
    return {"data": [data]}



import csv
from io import StringIO

import numpy as np
import pandas as pd
from wbcore.contrib.geography.models import Geography
from xlrd import xldate_as_datetime

from wbportfolio.models import Register

from .sylk import SYLK


def convert_string_to_number(string):
    try:
        return float(string.replace(" ", "").replace(",", ""))
    except ValueError:
        return 0.0


def parse(import_source):
    data = list()

    sylk_handler = SYLK()
    for line in [_line.decode("cp1252") for _line in import_source.file.open("rb").readlines()]:
        sylk_handler.parseline(line)

    buffer = StringIO()
    csvwriter = csv.writer(buffer, quotechar="'", delimiter=";", lineterminator="\n", quoting=csv.QUOTE_ALL)
    for line in sylk_handler.stream_rows():
        csvwriter.writerow(line)

    buffer.seek(0)
    content = buffer.read().replace('""', "")
    df = pd.read_csv(
        StringIO(content),
        sep=";",
        quotechar="'",
        usecols=[
            2,
            3,
            4,
            5,
            8,
            9,
            10,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
            21,
            22,
            23,
            24,
            25,
            26,
            27,
            28,
            29,
            30,
            32,
            33,
            34,
            36,
            42,
            43,
            46,
            47,
            48,
            49,
            50,
            55,
            56,
            57,
            58,
        ],
    )

    # Convert timestamps to json conform date strings
    df["ESTABLISHMENT_DATE"] = df["ESTABLISHMENT_DATE"].apply(lambda x: xldate_as_datetime(x, datemode=0).date())
    df["ESTABLISHMENT_DATE"] = df["ESTABLISHMENT_DATE"].apply(lambda x: x.strftime("%Y-%m-%d"))

    df = df.replace(np.nan, "", regex=True)
    # df = df.astype(str)
    df["GLOBAL_REGISTER_NUMBER"] = df["GLOBAL_REGISTER_NUMBER"].apply(lambda x: str(int(x)) if x else "")

    # Merge TextFields
    df["custodian_address"] = df[["ADRESS_LINE_1_4", "ADRESS_LINE_2_4", "ADRESS_LINE_3_4", "ADRESS_LINE_4_4"]].apply(
        lambda x: "<br />".join(x.dropna().astype(str)), axis=1
    )

    def concat_status(x):
        status = []

        if code := x["BLOCK_CODE"]:
            try:
                status.append(f"{int(code):04}")
            except ValueError:
                status.append(code)
        if desc := x["BLOCK_CODE_DESCRIPTION"]:
            status.append(desc)

        for n in range(1, 5):
            if code := x[f"BLOCK_CODE_{n}"]:
                try:
                    status.append(f"{int(code):04}")
                except ValueError:
                    status.append(code)
            if desc := x[f"DESCRIPTION_BC{n}"]:
                status.append(desc)

        return "<br />".join(status)

    df["status_message"] = df.apply(concat_status, axis=1)

    # Map Status
    def apply_status(x):
        if x.get("status_message", None) and x["ACTIVE_INACTIVE"] != "INACTIVE":
            return Register.RegisterStatus.WARNING

        mapping = {
            "ACTIVE": Register.RegisterStatus.ACTIVE,
            "INACTIVE": Register.RegisterStatus.INACTIVE,
        }
        return mapping[x["ACTIVE_INACTIVE"]]

    df["status"] = df.apply(apply_status, axis=1)

    # Remove not used columns
    del df["ADRESS_LINE_1_4"]
    del df["ADRESS_LINE_2_4"]
    del df["ADRESS_LINE_3_4"]
    del df["ADRESS_LINE_4_4"]
    del df["BLOCK_CODE"]
    del df["BLOCK_CODE_DESCRIPTION"]
    del df["BLOCK_CODE_1"]
    del df["DESCRIPTION_BC1"]
    del df["BLOCK_CODE_2"]
    del df["DESCRIPTION_BC2"]
    del df["BLOCK_CODE_3"]
    del df["DESCRIPTION_BC3"]
    del df["BLOCK_CODE_4"]
    del df["DESCRIPTION_BC4"]
    del df["ACTIVE_INACTIVE"]

    # Map Countries
    country_code_mapping = {
        country["code_2"]: country["id"] for country in Geography.countries.all().values("id", "code_2")
    }
    country_code_mapping["XX"] = country_code_mapping["ZZ"]
    country_title_mapping = {
        country["name"].lower(): country["id"] for country in Geography.countries.all().values("id", "name")
    }
    country_title_mapping["great-britain"] = Geography.countries.get(code_2="GB").id
    country_title_mapping["man (isle of)"] = Geography.countries.get(code_2="IM").id

    df["CITIZENSHIP"] = df["CITIZENSHIP"].apply(lambda x: country_code_mapping[x])
    df["RESIDENCE"] = df["RESIDENCE"].apply(lambda x: country_code_mapping[x])
    df["CPAYS"] = df["CPAYS"].apply(lambda x: country_code_mapping[x])
    df["COUNTRY"] = df["COUNTRY"].apply(lambda x: country_title_mapping[x.lower()])
    df["CITY"] = df["CITY"].str.title()
    # Map Investor Type
    investor_mapping = {
        "0006": Register.RegisterInvestorType.BANK,
        "0011": Register.RegisterInvestorType.NOMINEE,
        "0003": Register.RegisterInvestorType.GLOBAL,
        "0007": Register.RegisterInvestorType.NON_FINANCIAL_ENTITY,
    }
    df["INVESTOR_TYPE_CODE"] = df["INVESTOR_TYPE_CODE"].apply(lambda x: investor_mapping[f"{x:04}"])

    # Rename Columns
    df.columns = [
        "register_reference",
        "opened",
        "register_name_1",
        "register_name_2",
        "custodian_reference",
        "custodian_name_1",
        "custodian_name_2",
        "custodian_postcode",
        "custodian_city",
        "custodian_country",
        "investor_type",
        "global_register_reference",
        "citizenship",
        "residence",
        "external_register_reference",
        "sales_reference",
        "dealer_reference",
        "outlet_reference",
        "outlet_name",
        "outlet_address",
        "outlet_postcode",
        "outlet_country",
        "opened_reference_1",
        "opened_reference_2",
        "updated_reference_1",
        "updated_reference_1",
        "custodian_address",
        "status_message",
        "status",
    ]
    # Convert df to list of dicts
    data = list(df.T.to_dict().values())

    return {"data": data}



import json
from datetime import datetime
from io import BytesIO

from django.core.serializers.json import DjangoJSONEncoder
from pandas.tseries.offsets import BDay
from wbcore.contrib.io.backends import AbstractDataBackend, register
from wbfdm.models import Instrument

from wbportfolio.import_export.backends.utils import (
    get_timedelta_import_instrument_price,
)

from .mixin import DataBackendMixin


@register("Adjustment", provider_key="wbfdm", save_data_in_import_source=False, passive_only=False)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    DATE_LABEL = "adjustment_date"
    DEFAULT_MAPPING = {
        "adjustment_date": "date",
        "adjustment_factor": "factor",
        "cumulative_adjustment_factor": "cumulative_factor",
        "instrument_id": "instrument",
    }
    NONE_NULLABLE_FIELDS = list(DEFAULT_MAPPING.values())

    def get_default_queryset(self):
        return Instrument.objects.filter(is_investable_universe=True)

    def get_files(
        self,
        execution_time: datetime,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()
        start = kwargs.get("start", (execution_date - BDay(get_timedelta_import_instrument_price())).date())
        data = []
        for dict_dto in self.get_default_queryset().dl.adjustments(
            from_date=start,
            to_date=execution_date,
        ):
            if dict_dto["adjustment_factor"] is not None:
                data.append(dict_dto)
        if data:
            content_file = BytesIO()
            content_file.write(json.dumps(data, cls=DjangoJSONEncoder).encode())
            file_name = (
                f"adjustment_{start:%Y-%m-%d}-{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
            )
            yield file_name, content_file



from datetime import date
from typing import TYPE_CHECKING

import pytest
from django.db import IntegrityError
from wbcrm.models import Account

from wbportfolio.models import AccountReconciliation, Claim
from wbportfolio.models.products import Product
from wbportfolio.models.reconciliations.account_reconciliation_lines import (
    AccountReconciliationLine,
)

if TYPE_CHECKING:
    from wbcore.contrib.authentication.models import User

    from wbportfolio.factories import AccountReconciliationLineFactory, ClaimFactory


class TestAccountReconciliation:
    def test_str(self):
        account = Account(computed_str="Test Account")
        reconciliation = AccountReconciliation(account=account, reconciliation_date=date(2021, 1, 1))
        assert str(reconciliation) == "Test Account (01.01.2021)"

    def test_get_endpoint_basename(self):
        assert AccountReconciliation.get_endpoint_basename() == "wbportfolio:accountreconciliation"

    def test_get_representation_endpoint(self):
        assert (
            AccountReconciliation.get_representation_endpoint()
            == "wbportfolio:accountreconciliationrepresentation-list"
        )

    def test_get_representation_value_key(self):
        assert AccountReconciliation.get_representation_value_key() == "id"

    def test_get_representation_label_key(self):
        assert AccountReconciliation.get_representation_label_key() == "{{account}} {{reconciliation_date}}"

    @pytest.mark.django_db
    def test_create_for_date_and_account(self, claim_factory: "ClaimFactory", account: Account, user: "User"):
        for _ in range(10):
            claim_factory.create(account=account, date=date(2020, 12, 30), status=Claim.Status.APPROVED)

        AccountReconciliation.objects.create(
            reconciliation_date=date(2021, 1, 1),
            account=account,
            creator=user,
        )

        assert (
            AccountReconciliationLine.objects.count() == Claim.objects.distinct("product").count()
        ), "There should be as many lines as unique products for claims"


class TestAccountReconciliationLine:
    def test_str(self):
        account = Account(computed_str="Test Account")
        reconciliation = AccountReconciliation(account=account, reconciliation_date=date(2021, 1, 1))
        product = Product(computed_str="Test Product")
        line = AccountReconciliationLine(reconciliation=reconciliation, product=product, shares=100)

        assert str(line) == "Test Account (01.01.2021): Test Product 100 shares"

    def test_get_endpoint_basename(self):
        assert AccountReconciliationLine.get_endpoint_basename() == "wbportfolio:accountreconciliationline"

    def test_representation_endpoint(self):
        assert (
            AccountReconciliationLine.get_representation_endpoint()
            == "wbportfolio:accountreconciliationlinerepresentation-list"
        )

    def test_representation_value_key(self):
        assert AccountReconciliationLine.get_representation_value_key() == "id"

    def test_representation_label_key(self):
        assert AccountReconciliationLine.get_representation_label_key() == "{{reconciliation}}"

    def test_save_calculate_nominal_value(self, mocker):
        mocker.patch("django.db.models.Model.save")
        product = Product(share_price=100)
        line = AccountReconciliationLine(product=product, shares=100, nominal_value=None)
        line.save()

        assert line.nominal_value == 100 * 100

    def test_save_calculate_shares(self, mocker):
        mocker.patch("django.db.models.Model.save")
        product = Product(share_price=100)
        line = AccountReconciliationLine(product=product, nominal_value=10000, shares=None)
        line.save()

        assert line.shares == 100

    def test_save_calculate_nominal_value_external(self, mocker):
        mocker.patch("django.db.models.Model.save")
        product = Product(share_price=100)
        line = AccountReconciliationLine(product=product, shares_external=100, nominal_value_external=None)
        line.save()

        assert line.nominal_value_external == 100 * 100

    def test_save_calculate_shares_external(self, mocker):
        mocker.patch("django.db.models.Model.save")
        product = Product(share_price=100)
        line = AccountReconciliationLine(product=product, nominal_value_external=10000, shares_external=None)
        line.save()

        assert line.shares_external == 100

    @pytest.mark.django_db
    def test_unique_constraint(
        self,
        account_reconciliation: AccountReconciliation,
        product: Product,
        account_reconciliation_line_factory: "AccountReconciliationLineFactory",
    ):
        account_reconciliation_line_factory.create(reconciliation=account_reconciliation, product=product)
        with pytest.raises(IntegrityError):
            account_reconciliation_line_factory.create(reconciliation=account_reconciliation, product=product)

    @pytest.mark.django_db
    def test_currency_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert (
            AccountReconciliationLine.objects.all().annotate_currency().first().currency
            == account_reconciliation_line.product.currency.symbol
        )

    @pytest.mark.django_db
    def test_assets_under_management_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert (
            AccountReconciliationLine.objects.all().annotate_assets_under_management().first().assets_under_management
            == account_reconciliation_line.shares * account_reconciliation_line.price
        )

    @pytest.mark.django_db
    def test_assets_under_management_external_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert (
            AccountReconciliationLine.objects.all()
            .annotate_assets_under_management_external()
            .first()
            .assets_under_management_external
            == account_reconciliation_line.shares_external * account_reconciliation_line.price
        )

    @pytest.mark.django_db
    @pytest.mark.parametrize("account_reconciliation_line__shares", [100])
    @pytest.mark.parametrize("account_reconciliation_line__shares_external", [101])
    def test_is_not_equal_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert not AccountReconciliationLine.objects.all().annotate_is_equal().first().is_equal

    @pytest.mark.django_db
    @pytest.mark.parametrize("account_reconciliation_line__shares", [100])
    @pytest.mark.parametrize("account_reconciliation_line__shares_external", [100])
    def test_is_equal_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert AccountReconciliationLine.objects.all().annotate_is_equal().first().is_equal

    @pytest.mark.django_db
    def test_shares_diff_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert (
            AccountReconciliationLine.objects.all().annotate_shares_diff().first().shares_diff
            == account_reconciliation_line.shares_external - account_reconciliation_line.shares
        )

    @pytest.mark.django_db
    def test_pct_diff_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert (
            AccountReconciliationLine.objects.all().annotate_pct_diff().first().pct_diff
            == (account_reconciliation_line.shares_external - account_reconciliation_line.shares)
            / account_reconciliation_line.shares
        )

    @pytest.mark.django_db
    def test_nominal_value_diff_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert (
            AccountReconciliationLine.objects.all().annotate_nominal_value_diff().first().nominal_value_diff
            == account_reconciliation_line.nominal_value_external - account_reconciliation_line.nominal_value
        )

    @pytest.mark.django_db
    def test_assets_under_management_diff_annotation(self, account_reconciliation_line: AccountReconciliationLine):
        assert (
            AccountReconciliationLine.objects.all()
            .annotate_assets_under_management()
            .annotate_assets_under_management_external()
            .annotate_assets_under_management_diff()
            .first()
            .assets_under_management_diff
            == (account_reconciliation_line.shares_external * account_reconciliation_line.price)
            - (account_reconciliation_line.shares * account_reconciliation_line.price)
        )



import json

import numpy as np
import pandas as pd
from wbcore.contrib.currency.models import CurrencyFXRates

from wbportfolio.import_export.utils import extract_exchange_ticker

BASE_MAPPING = {
    "instrument": "underlying_quote__name",
    "ric": "underlying_quote__refinitiv_identifier_code",
    "bbTicker": "underlying_quote__ticker",
    "isin": "underlying_quote__isin",
    "sedol": "underlying_quote__sedol",
    "valoren": "underlying_quote__valoren",
    "assetClass": "underlying_quote__instrument_type",
    "percentageWeight": "weighting",
    "shares": "initial_shares",
    "localClose": "initial_price",
    "currency": "currency__key",
    "fxMultiplier": "initial_currency_fx_rate",
}

INSTRUMENT_MAPPING = {
    "EQUITY": {},
    "INDEX": {},
    "ETF": {},
    "UBS_SECURITY": {},
    "FUTURE": {"contracts": "contracts"},
    "CASH": {"currencyAmount": "initial_shares"},
    "CASH_COMP": {
        "currencyAmount": "initial_shares",
    },
    "BOND": {"creditRating": "credit_rating", "maturityDate": "maturity_date"},
    "OPTION": {
        "description": "description",
        "strike": "strike",
        "sizeOfContract": "size_of_contract",
        "exerciseStyle": "exercise_style",
        "optionType": "option_type",
        "expiryDate": "expiry_date",
        "underlying__description": "underlying__description",
        "underlying__ric": "underlying__refinitiv_identifier_code",
        "underlying__bbTicker": "underlying__ticker",
        "underlying__assetClass": "underlying__instrument_type",
    },
}

INSTRUMENT_TYPE_MAP = {
    "EQUITY": "equity",
    "INDEX": "index",
    "ETF": "etf",
    "UBS_SECURITY": "equity",
    "FUTURE": "future",
    "CASH": "cash",
    "CASH_COMP": "cash",
    "BOND": "bond",
    "OPTION": "option",
}


def parse(import_source):
    content = json.load(import_source.file)
    data = list()
    if (
        (constituents := content.get("constituents", None))
        and (val_date := content.get("asOfDate", None))
        and (amc := content.get("amc", None))
    ):
        for asset_class, asset_data in constituents.items():
            if asset_class in INSTRUMENT_MAPPING.keys():
                asset_class_map = INSTRUMENT_MAPPING[asset_class]
                _map = {**asset_class_map, **BASE_MAPPING}
                df = pd.json_normalize(asset_data, sep="__")
                df = df.replace([np.inf, -np.inf, np.nan], None)
                df = df.rename(columns=_map).dropna(how="all", axis=1)
                df["weighting"] = df["weighting"] / 100.0
                df["date"] = val_date
                df["portfolio__instrument_type"] = "product"
                df["portfolio__isin"] = amc["isin"]
                df["underlying_quote__currency__key"] = df["currency__key"]
                df["underlying_quote__instrument_type"] = df.underlying_quote__instrument_type.apply(
                    lambda x: INSTRUMENT_TYPE_MAP[x]
                )
                if asset_class == "CASH":
                    df["initial_price"] = 1.0
                    df["initial_currency_fx_rate"] = (
                        df["currency__key"]
                        .apply(
                            lambda x: CurrencyFXRates.objects.filter(currency__key=x, date__lte=val_date)
                            .latest("date")
                            .value
                        )
                        .astype(float)
                    )
                df["is_estimated"] = False
                if "underlying_quote__ticker" in df.columns:
                    df["underlying_quote__exchange"] = df["underlying_quote__ticker"].apply(
                        lambda x: {"bbg_exchange_codes": extract_exchange_ticker(x)[1]}
                    )
                    df["underlying_quote__ticker"] = df["underlying_quote__ticker"].apply(
                        lambda x: extract_exchange_ticker(x)[0]
                    )
                if "underlying_quote__exchange" in df.columns:
                    df["exchange"] = df["underlying_quote__exchange"]
                data.extend(df.to_dict("records"))
    return {"data": data}



from django.contrib import admin

from .models import (
    AssetAllocation,
    AssetAllocationType,
    CompanyPortfolioData,
    GeographicFocus,
)


@admin.register(CompanyPortfolioData)
class CompanyPortfolioDataModelAdmin(admin.ModelAdmin):
    pass


@admin.register(AssetAllocationType)
class AssetAllocationTypeModelAdmin(admin.ModelAdmin):
    pass


@admin.register(AssetAllocation)
class AssetAllocationModelAdmin(admin.ModelAdmin):
    pass


@admin.register(GeographicFocus)
class GeographicFocusModelAdmin(admin.ModelAdmin):
    pass



# Generated by Django 5.0.6 on 2024-08-15 13:59

from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0064_alter_portfolio_managers_portfolio_is_tracked_and_more"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="portfolio",
            managers=[],
        ),
        migrations.AddField(
            model_name="claim",
            name="as_shares",
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="nominal_value",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The nominal value computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="nominal_value_external",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The nominal value externally provided through the Account holder. Initially set to the number of shares computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="shares",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The number of shares computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="accountreconciliationline",
            name="shares_external",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The number of shares externally provided through the Account holder. Initially set to the number of shares computed through the Workbench",
                max_digits=18,
            ),
        ),
        migrations.AlterField(
            model_name="claim",
            name="date",
            field=models.DateField(verbose_name="Trade Date"),
        ),
        migrations.AlterField(
            model_name="claim",
            name="reference",
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name="Additional Reference"),
        ),
    ]



from django.contrib import admin

from wbportfolio.admin.transactions import TransactionModelAdmin
from wbportfolio.models import DividendTransaction


@admin.register(DividendTransaction)
class DividendAdmin(TransactionModelAdmin):
    fieldsets = (
        ("Transactions Information", TransactionModelAdmin.fieldsets[0][1]),
        (
            "Dividend Information",
            {"fields": (("retrocession",),)},
        ),
    )



import re
from datetime import datetime
from io import BytesIO

import numpy as np
import pandas as pd

from wbportfolio.models import Product

FIELD_MAP = {
    "CCY": "currency__key",
    "CURRENT PRICE": "initial_price",
    "FX": "initial_currency_fx_rate",
    "TOTAL UNITS": "initial_shares",
    "TYPE": "underlying_quote__instrument_type",
    "NAME": "underlying_quote__name",
    "ISIN": "underlying_quote__isin",
    "WEIGHT (%)": "weighting",
}


def parse(import_source):
    data = list()

    isin_regex = "([A-Z]{2}[A-Z0-9]{9}[0-9]{1})"

    dict_df = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl", sheet_name=None)

    for sheet_name, df in dict_df.items():
        if len(re.findall(isin_regex, sheet_name)) > 0:
            valuation_date = datetime.strptime(df.iloc[0, 1].split("AMC OVERVIEW REPORT AS OF ")[-1], "%d.%m.%Y")
            if valuation_date.weekday() not in [5, 6]:
                xx, yy = np.where(df == "COMPOSITION DATA")
                df = df.iloc[xx[0] + 1 :, yy[0] :]
                df = df.rename(columns=df.iloc[0]).drop(df.index[0]).dropna(how="all").rename(columns=FIELD_MAP)

                df = df[
                    (df["underlying_quote__instrument_type"].isin(["CASH", "SHARE"]))
                    & (df["N"].astype("str").str.isnumeric())
                ]

                equities_index = df["underlying_quote__instrument_type"] == "SHARE"
                df.loc[~equities_index, "initial_shares"] = df.loc[~equities_index, "TOTAL VALUE"]
                df.loc[~equities_index, "initial_price"] = 1.0
                df.initial_currency_fx_rate = df.initial_currency_fx_rate.fillna(1.0)
                # df["weighting"] = df.initial_currency_fx_rate * df.initial_price * df.initial_shares
                # df["weighting"] = df.weighting / df.weighting.sum()
                product = Product.objects.get(isin=sheet_name)
                for position in df.to_dict("records"):
                    if position["underlying_quote__instrument_type"] == "CASH":
                        ticker = exchange = "CASH"
                    else:
                        ticker, exchange, _ = str(position["BBG TICKER / IDENTIFIER"]).split(" ")

                    if position["underlying_quote__instrument_type"] == "SHARE":
                        underlying_quote = {
                            "ticker": ticker,
                            "exchange": {"bbg_exchange_codes": exchange},
                            "name": position["underlying_quote__name"],
                            "currency__key": position["currency__key"],
                            "instrument_type": "equity",
                            "isin": position["underlying_quote__isin"],
                        }
                    else:
                        underlying_quote = {"currency__key": position["currency__key"], "instrument_type": "cash"}
                    data.append(
                        {
                            "underlying_quote": underlying_quote,
                            "currency__key": position["currency__key"],
                            "is_estimated": False,
                            "date": valuation_date.strftime("%Y-%m-%d"),
                            "asset_valuation_date": valuation_date.strftime("%Y-%m-%d"),
                            "initial_price": position["initial_price"],
                            "initial_currency_fx_rate": position["initial_currency_fx_rate"],
                            "initial_shares": position["initial_shares"],
                            "weighting": position["weighting"],
                            "exchange": {"bbg_exchange_codes": exchange},
                            "portfolio": {"instrument_type": "product", "id": product.id},
                        }
                    )

    return {"data": data}



import numpy as np
import pandas as pd

FIELD_MAP = {
    "#AssetIdentifier": "identifier",
    "AssetCurrency": "currency__key",
    "AssetValor": "valoren",
    "ISIN": "isin",
    "SymbolTK": "ticker",
    "OperatingMic": "exchange",
    "AssetType": "instrument_type",
    "AssetText": "name",
}

ASSET_TYPE_MAP = {
    "EQUADR": "equity",
    "EQUBEA": "equity",
    "EQUREG": "equity",
    "FNDETF": "etf",
    "OTHOTH": np.nan,
    "STROTH": "etf",
    "STRPRT": "etf",
}


def parse(import_source):
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=";")
    df = df.rename(columns=FIELD_MAP)
    df = df.replace([np.inf, -np.inf, np.nan], None)
    df["instrument_type"] = df["instrument_type"].apply(lambda x: ASSET_TYPE_MAP[x])
    df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
    df["exchange"] = df["exchange"].apply(lambda x: {"operating_mic_code": x})
    df = df.dropna(subset="instrument_type")
    return {"data": df.to_dict("records")}



from contextlib import suppress
from datetime import datetime

import pandas as pd

from wbportfolio.models import Portfolio


def parse(import_source):
    csv_file = import_source.file.open()

    df = pd.read_csv(csv_file)

    def get_portfolio_id(row) -> int | None:
        with suppress(Portfolio.DoesNotExist):
            return Portfolio.objects.get(instruments__children__isin__in=[row["Isin"]]).pk

    df = df[~df["Isin"].isnull()]
    df["Trade date"] = df["Trade date"].ffill()
    df["portfolio"] = df.apply(get_portfolio_id, axis=1)
    df = df[["Trade date", "portfolio", "Net Ref ccy"]]
    df = df.groupby(["Trade date", "portfolio"]).sum().to_dict()

    data = []
    for keys, cash_flow_forecast in df["Net Ref ccy"].items():
        data.append(
            {
                "value_date": datetime.strptime(keys[0], "%d/%m/%Y").strftime("%Y-%m-%d"),
                "portfolio": keys[1],
                "cash_flow_forecast": cash_flow_forecast,
                "pending": True,
            }
        )

    return {
        "data": data,
    }



import factory
from faker import Faker

from wbportfolio.models import Fees

from .transactions import TransactionFactory

faker = Faker()


class FeesFactory(TransactionFactory):
    class Meta:
        model = Fees

    transaction_subtype = factory.Faker("random_element", elements=[x[0] for x in Fees.Type.choices])
    linked_product = factory.SubFactory("wbportfolio.factories.ProductFactory")



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AssetPositionPandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbportfolio:assetpositiongroupby-list", request=self.request)


class AggregatedAssetPositionLiquidityEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbportfolio:aggregatedassetpositionliquidity-list", request=self.request)



from contextlib import suppress
from datetime import date, datetime, time
from decimal import Decimal

from celery import shared_task
from django.contrib import admin
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateRangeField, RangeOperators
from django.db import models
from django.db.models import (
    BooleanField,
    Case,
    DecimalField,
    ExpressionWrapper,
    F,
    OuterRef,
    Subquery,
    Value,
    When,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import pre_save
from django.dispatch import receiver
from pandas.tseries.offsets import BDay
from wbcore.contrib.ai.llm.config import add_llm_prompt
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcore.signals import pre_merge
from wbcore.utils.enum import ChoiceEnum
from wbcrm.models.accounts import Account
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbfdm.models.instruments.instruments import InstrumentManager, InstrumentType

from wbportfolio.models.llm.wbcrm.analyze_relationship import get_holding_prompt
from wbportfolio.models.portfolio_relationship import InstrumentPortfolioThroughModel

from .mixins.instruments import PMSInstrument, PMSInstrumentAbstractModel


class TypeOfReturn(ChoiceEnum):
    TOTAL_RETURN = "Total return"
    YIELD = "Yield"
    PRICE_RETURN = "Price return"
    ABSOLUT_RETURN = "Absolut return"


class AssetClass(ChoiceEnum):
    EQUITY = "Equity"
    FIXED_INCOME = "Fixed income"
    PRIVATE_EQUITY = "Private equity"
    DERIVATIVE = "Derivative"
    CASH = "Cash"
    CRYPTOCURRENCY = "Cryptocurrency"
    COMMODITY = "Commodity"
    HEDGE_FUND = "Hedge Fund"
    MULTI_ASSET = "Multi Asset"


class LegalStructure(ChoiceEnum):
    AMC = "Actively Managed Certificate"
    SICAV_UCITS = "SICAV (UCITS)"
    PLC_UCITS = "PLC (UCITS)"
    SICAV_AIF = "SICAV (AIFs)"
    SCF_UCITS = "Swiss Contractual Fund (UCITS)"
    FCP_AIF = "FCP (AIFs)"
    PLC_AIF = "PLC (AIFs)"
    SCF_AIF = "Swiss Contractual Fund (AIF) "


class InvestmentIndex(ChoiceEnum):
    LONG = "Long"
    SHORT = "Short"
    LONG_SHORT = "Long/Short"


class Liquidy(ChoiceEnum):
    DAILY = "Daily"
    WEEKLY = "Weekly"
    MONTHLY = "Monthly"
    QUATERLY = "Quaterly"
    YEARLY = "Yearly"


class DefaultProductManager(InstrumentManager):
    def get_queryset(self):
        today = date.today()
        return (
            super()
            .get_queryset()
            .annotate(
                value_date=Case(
                    When(last_valuation_date__isnull=False, then=F("last_valuation_date")),
                    default=Value(today),
                    output_field=models.DateField(),
                ),
                is_invested=models.Exists(
                    InstrumentPortfolioThroughModel.objects.filter(
                        instrument=models.OuterRef("pk"),
                        portfolio__invested_timespan__startswith__lte=OuterRef("value_date"),
                        portfolio__invested_timespan__endswith__gt=OuterRef("value_date"),
                    )
                ),
                current_bank_fees=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.BANK,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("percent")[:1]
                    ),
                    Decimal(0),
                ),
                current_management_fees=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.MANAGEMENT,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("percent")[:1]
                    ),
                    Decimal(0),
                ),
                current_performance_fees=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.PERFORMANCE,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("percent")[:1]
                    ),
                    Decimal(0),
                ),
                current_performance_fees_vat_deduction=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.PERFORMANCE,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("vat_deduction")[:1]
                    ),
                    Decimal(0),
                ),
                current_gross_performance_fees_percent=F("current_performance_fees")
                + F("current_performance_fees_vat_deduction"),
                current_total_issuer_fees=F("current_bank_fees") + F("current_management_fees"),
            )
        )


class FeeProductPercentage(models.Model):
    class Type(models.TextChoices):
        MANAGEMENT = "MANAGEMENT", "Management"
        PERFORMANCE = "PERFORMANCE", "Performance"
        BANK = "BANK", "Bank"

    product = models.ForeignKey("wbportfolio.Product", related_name="fees_percentages", on_delete=models.CASCADE)
    type = models.CharField(max_length=16, default=Type.MANAGEMENT, choices=Type.choices)
    vat_deduction = models.DecimalField(
        max_digits=6,
        decimal_places=6,
        default=Decimal(0),
        verbose_name="VAT Deduction",
        help_text="The VAT deducted from this fees percentage",
    )
    percent = models.DecimalField(max_digits=6, decimal_places=6)
    timespan = DateRangeField()

    class Meta:
        verbose_name = "Fees Percentage"
        verbose_name_plural = "Fees Percentage"
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_product_fees_type",
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("product", RangeOperators.EQUAL),
                    ("type", RangeOperators.EQUAL),
                ],
            ),
        ]

    @property
    def net_percent(self) -> Decimal:
        return self.percent

    @property
    def gross_percent(self) -> Decimal:
        return self.percent + self.vat_deduction


class Product(PMSInstrumentAbstractModel):
    share_price = models.PositiveIntegerField(
        default=100,
        verbose_name="Share Price",
        help_text="The initial share price that is used to calculate the nominal value of a product.",
    )
    initial_high_water_mark = models.PositiveIntegerField(
        default=100,
        verbose_name="Initial High Water Mark",
        help_text="Initial High Water Mark",
    )

    bank = models.ForeignKey(
        "directory.Company",
        related_name="issues_products",
        on_delete=models.PROTECT,
        verbose_name="Bank",
        help_text="The Bank that holds the product. A company from the CRM.",
    )

    white_label_customers = models.ManyToManyField(
        "directory.Entry",
        related_name="white_label_customer_products",
        blank=True,
        verbose_name="White Label Customers",
        help_text="Specifies whether a product is a white label product or not. If at least one customer is specified, this product becomes a white label product and can only be seen by the respected customers.",
    )

    default_account = models.ForeignKey(
        "wbcrm.Account",
        related_name="default_product_accounts",
        limit_choices_to=models.Q(is_terminal_account=True),
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name="Default Account",
        help_text="If a default Account is set, then all newly created tradesare automatically matched to this sub account.",
    )

    fee_calculation = models.ForeignKey(
        "wbportfolio.FeeCalculation",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="products",
        verbose_name="Fee Calculation Method",
    )

    termsheet = models.FileField(max_length=256, null=True, blank=True, upload_to="portfolio/product/termsheets")

    class Dividend(models.TextChoices):
        CAPITALISATION = "CAPITALISATION", "Capitalisation"
        DISTRIBUTION = "DISTRIBUTION", "Distribution"

    type_of_return = models.CharField(
        max_length=16,
        default=TypeOfReturn.TOTAL_RETURN.name,
        choices=TypeOfReturn.choices(),
        verbose_name="Type of Return",
    )
    asset_class = models.CharField(
        max_length=16,
        default=AssetClass.EQUITY.name,
        choices=AssetClass.choices(),
        verbose_name="Asset Class",
    )
    legal_structure = models.CharField(
        max_length=16,
        default=LegalStructure.AMC.name,
        choices=LegalStructure.choices(),
        verbose_name="Legal Structure",
    )
    investment_index = models.CharField(
        max_length=16,
        default=InvestmentIndex.LONG.name,
        choices=InvestmentIndex.choices(),
        verbose_name="Long/Short Strategy",
    )
    liquidity = models.CharField(
        max_length=16,
        default=Liquidy.DAILY.name,
        choices=Liquidy.choices(),
        verbose_name="Liquidity",
    )

    jurisdiction = models.ForeignKey(
        "geography.Geography",
        related_name="products",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Country on which this product is issued and bonded by law",
        help_text="The country of jurisdiction",
        limit_choices_to={"level": 1},
    )

    dividend = models.CharField(
        max_length=16,
        default=Dividend.CAPITALISATION,
        choices=Dividend.choices,
        verbose_name="Type of Dividends",
    )

    minimum_subscription = models.IntegerField(
        default=100,
        help_text="Minimum subscription amount allowed to invest in the fund",
    )

    cut_off_time = models.TimeField(default=time(13, 0, 0))
    external_webpage = models.URLField(blank=True, null=True)

    bank_account = models.ForeignKey(
        to="directory.BankingContact",
        related_name="wbportfolio_products",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
    )

    objects = DefaultProductManager()

    class Meta:
        verbose_name = "Product"
        verbose_name_plural = "Products"

    def pre_save(self):
        super().pre_save()
        self.instrument_type = InstrumentType.PRODUCT
        if "market_data" not in self.dl_parameters:
            # we default to the internal dataloader
            self.dl_parameters["market_data"] = {
                "path": "wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader"
            }

        self.is_managed = True

    def get_title(self):
        if self.parent:
            return f"{self.parent.name} ({self.name})"
        return self.name

    def compute_str(self):
        computed_str = f"{self.get_title()} - {self.bank.name} - {self.currency.key}"
        if self.id:
            fees = (self.bank_fees + self.management_fees) * 100
            computed_str += f" - {fees:.2f} % Mgmt Fees"
        computed_str += f" ({self.isin})"
        return computed_str

    @property
    def white_label_product(self):
        return self.white_label_customers.all().count() > 0

    @property
    def urlify_title(self):
        url = self.name + "-" + self.bank.name + "-" + self.isin + "-" + self.currency.key
        return url.lower()

    @property
    @admin.display(description="Invested")
    def _is_invested(self) -> bool:
        """
        Return True if the associated asset portfolio is invested
        """
        return getattr(self, "is_invested", self.portfolio and self.portfolio.is_invested_at_date(date.today()))

    @property
    def group(self):  # for backward compatibility
        from wbportfolio.models.product_groups import ProductGroup

        if self.parent:
            with suppress(ProductGroup.DoesNotExist):
                return ProductGroup.objects.get(id=self.parent.id)
        return None

    def get_fees_percent(self, val_date: date, fee_type: FeeProductPercentage.Type, net: bool = True) -> Decimal:
        try:
            fee = FeeProductPercentage.objects.get(
                product=self, type=fee_type, timespan__startswith__lte=val_date, timespan__endswith__gt=val_date
            )
            percent = fee.percent
            if not net:
                percent += fee.vat_deduction
            return percent
        except FeeProductPercentage.DoesNotExist:
            return Decimal(0)

    @property
    def bank_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.BANK)

    @property
    def management_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.MANAGEMENT)

    @property
    def performance_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.PERFORMANCE)

    @property
    def gross_performance_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.PERFORMANCE, net=False)

    def get_high_water_mark(self, val_date: date) -> Decimal:
        """Returns the high water mark of a instrument

        The high water mark is the highest previously reached net price

        Returns:
            decimal.Decimal -- High Water Mark
        """
        latest_high_water_mark = (
            self.valuations.filter(date__lt=val_date).aggregate(high_water_mark=models.Max(F("net_value")))[
                "high_water_mark"
            ]
            or self.share_price
        )
        return max(latest_high_water_mark, self.initial_high_water_mark)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:product"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:productrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}} ({{bank_name}})"

    @classmethod
    def get_last_valid_active_product_date(cls):
        current_date = (datetime.today() - BDay(1)).date()
        filter_date = current_date
        for product in Product.active_objects.all():
            if product.valuations.filter(date__lte=current_date).exists():
                last_valid_date = product.valuations.filter(date__lte=current_date).latest("date").date
                if last_valid_date < filter_date:
                    filter_date = last_valid_date
        return filter_date

    @classmethod
    def get_products(cls, person, base_qs: models.QuerySet | None = None):
        if base_qs is None:
            base_qs = cls.objects.all()
        if person.is_internal or person.user_account.is_superuser:
            return base_qs.all()
        return base_qs.filter(
            models.Q(white_label_customers__isnull=True)
            | models.Q(white_label_customers__in=[person, *person.employers.all()])
        ).distinct()

    @classmethod
    def subquery_is_white_label_product(cls, product_pk_name="pk"):
        return Coalesce(
            models.Subquery(
                Entry.objects.filter(white_label_customer_products__pk=models.OuterRef(product_pk_name))
                .values("white_label_customer_products")
                .annotate(
                    num_white_label_customers=models.Count("*"),
                    is_white_label_product=models.Case(
                        models.When(num_white_label_customers__gt=0, then=True),
                        output_field=models.BooleanField(),
                    ),
                )
                .values("is_white_label_product")[:1]
            ),
            False,
            output_field=models.BooleanField(),
        )

    @classmethod
    def annotate_last_aum(cls, qs, val_date=None, date_key="last_valuation_date"):
        if val_date:
            fx_rate_subquery = CurrencyFXRates.get_fx_rates_subquery(
                val_date, currency="currency", lookup_expr="exact"
            )
        else:
            fx_rate_subquery = CurrencyFXRates.get_fx_rates_subquery(
                date_key, currency="currency", lookup_expr="exact"
            )

        qs = InstrumentPrice.annotate_sum_shares(qs, val_date, date_key=date_key).annotate(
            net_value=Subquery(
                InstrumentPrice.objects.filter(
                    calculated=False, instrument=OuterRef("pk"), date=OuterRef("last_valuation_date")
                ).values("net_value")[:1]
            ),
            assets_under_management=Coalesce(F("sum_shares") * F("net_value"), Decimal(0)),
            fx_rate=Coalesce(
                fx_rate_subquery,
                Decimal(1.0),
            ),
            assets_under_management_usd=ExpressionWrapper(
                F("assets_under_management") * F("fx_rate"),
                output_field=DecimalField(),
            ),
            is_white_label=Case(
                When(white_label_customers__isnull=True, then=Value(False)),
                default=Value(True),
                output_field=BooleanField(),
            ),
        )
        return qs.distinct()


@shared_task(queue="portfolio")
def update_outstanding_shares_as_task(instrument_id: int, clear: bool = False):
    instrument = PMSInstrument.objects.get(id=instrument_id)
    instrument.update_outstanding_shares(clear=clear)


@receiver(pre_merge, sender="wbcrm.Account")
def handle_pre_merge_account_for_product(sender: models.Model, merged_object: Account, main_object: Account, **kwargs):
    """
    Simply reassign the product's default account linked to the merged account to the main account
    """
    Product.objects.filter(default_account=merged_object).update(default_account=main_object)


@receiver(pre_save, sender="wbfdm.Instrument")
def pre_save_instrument(sender, instance, raw, **kwargs):
    if not raw and instance.id:
        # Remove duplicates if existings
        instance.old_isins = list(set(instance.old_isins))
        pre_instance = sender.objects.get(id=instance.id)
        if (
            pre_instance.isin
            and instance.isin
            and pre_instance.isin != instance.isin
            and pre_instance.isin not in instance.old_isins
        ):
            instance.old_isins = [*instance.old_isins, pre_instance.isin]
        if pre_instance.currency != instance.currency:
            # currency has changed we need to recompute currency_fx_rate
            instance.prices.annotate(
                new_currency_fx_rate_to_usd_id=models.Subquery(
                    CurrencyFXRates.objects.filter(currency=instance.currency, date=models.OuterRef("date")).values(
                        "id"
                    )[:1]
                )
            ).update(currency_fx_rate_to_usd=models.F("new_currency_fx_rate_to_usd_id"))
            instance.assets.annotate(
                new_currency_fx_rate_instrument_to_usd_id=models.Subquery(
                    CurrencyFXRates.objects.filter(currency=instance.currency, date=models.OuterRef("date")).values(
                        "id"
                    )[:1]
                )
            ).update(currency_fx_rate_instrument_to_usd=models.F("new_currency_fx_rate_instrument_to_usd_id"))


@receiver(add_llm_prompt, sender="wbcrm.Account")
def add_performance_to_account_heat(sender, instance, key, **kwargs):
    if key == "analyze_relationship":
        return get_holding_prompt(instance)
    return []



import datetime
from decimal import Decimal

import pytest
from faker import Faker
from wbfdm.models import Instrument

from wbportfolio.models import Product, Trade

from .utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestProductModel(PortfolioTestMixin):
    def test_init(self, product):
        assert product.id is not None

    def test_white_label(self, white_label_product):
        assert white_label_product.white_label_product

    def test_total_value(self, product_factory, trade_factory, instrument_price_factory):
        product = product_factory.create()
        t1 = trade_factory.create(
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=2),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        t2 = trade_factory.create(
            shares=50,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        t3 = trade_factory.create(  # noise
            shares=10,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        p1 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t1.transaction_date
        )
        p2 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t2.transaction_date
        )
        p3 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t3.transaction_date
        )
        assert product.get_total_aum_usd(p1.date) == pytest.approx(Decimal(0), rel=Decimal("1e-4"))
        assert product.get_total_aum_usd(p2.date) == pytest.approx(
            Decimal(p2.net_value * (t1.shares)), rel=Decimal("1e-4")
        )
        assert product.get_total_aum_usd(p3.date) == pytest.approx(
            Decimal(p3.net_value * (t1.shares + t2.shares)), rel=Decimal("1e-4")
        )

    def test_total_shares_with_date(self, product_factory, trade_factory):
        date1 = datetime.date(2010, 1, 1)
        date2 = datetime.date(2010, 1, 2)
        product = product_factory.create()
        sub1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date1 - datetime.timedelta(days=1),
            value_date=date1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        sub2 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date2 - datetime.timedelta(days=1),
            value_date=date2,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        red1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date1 - datetime.timedelta(days=1),
            value_date=date1,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        red2 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date2 - datetime.timedelta(days=1),
            value_date=date2,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        assert product.total_shares(date1 - datetime.timedelta(days=1)) == 0
        assert product.total_shares(date1) == red1.shares + sub1.shares
        assert product.total_shares(date2) == red1.shares + red2.shares + sub1.shares + sub2.shares

    def test_nominal_value_with_date(self, product_factory, trade_factory):
        product = product_factory.create()
        trade_factory.create_batch(
            5,
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create_batch(
            5,
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 2) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        assert product.nominal_value(datetime.date(2010, 1, 1)) == (500 * product.share_price)
        assert product.nominal_value(datetime.date(2010, 1, 2)) == (1000 * product.share_price)

    def test_get_title(self, product, user):
        assert product.get_title()

    def test_compute_str(self, product, user):
        assert product.compute_str()

    def test_urlify_title(self, product):
        assert product.name.lower() in product.urlify_title

    def test_get_products_internal_user(self, product, white_label_product, internal_user_factory):
        assert set(Product.get_products(internal_user_factory.create().profile)) == {product, white_label_product}

    def test_get_products_normal_user(self, product_factory, user, company):
        public_product = product_factory.create()
        assert set(Product.get_products(user.profile)) == {public_product}

        white_label_product_profile = product_factory.create()
        white_label_product_profile.white_label_customers.add(user.profile)

        white_label_product_employer = product_factory.create()
        user.profile.employers.add(company)
        white_label_product_employer.white_label_customers.add(company)

        assert set(Product.get_products(user.profile)) == {
            white_label_product_employer,
            white_label_product_profile,
            public_product,
        }

    def test_subquery_is_white_label_product(self, product, white_label_product):
        product_queryset = Product.objects.all().annotate(is_white_label=Product.subquery_is_white_label_product())
        # for product in product_queryset:
        #     print(product, product.is_white_label)

        # print(product_queryset.filter(is_white_label=True).count())

        assert product_queryset.filter(is_white_label=True).count() == 1

    def test_annotate_aum(
        self, portfolio, product_factory, instrument_price_factory, trade_factory, currency_fx_rates_factory
    ):
        p1 = product_factory.create()
        d = datetime.date(2021, 1, 1)
        c1 = trade_factory.create(
            portfolio=p1.portfolio,
            underlying_instrument=p1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            value_date=d,
            transaction_date=d,
        )
        v1 = instrument_price_factory.create(instrument=p1, date=d, outstanding_shares=c1.shares)
        fx_rate = p1.currency.fx_rates.get(date=d)
        qs = Product.annotate_last_aum(Product.objects.all(), val_date=d)
        p1.update_last_valuation_date()
        assert float(qs.first().assets_under_management_usd) == pytest.approx(
            float(v1.net_value) * float(c1.shares) / float(fx_rate.value)
        )

    def test_account_merging(self, account_factory, product):
        base_account = account_factory.create()
        merged_account = account_factory.create()
        product.default_account = merged_account
        product.save()
        base_account.merge(merged_account)
        product.refresh_from_db()
        assert product.default_account == base_account

    def test_high_water_mark(self, product, instrument_price_factory):
        initial_price = instrument_price_factory.create(instrument=product, net_value=100)
        p1 = instrument_price_factory.create(instrument=product, net_value=200)
        p2 = instrument_price_factory.create(instrument=product, net_value=100)
        p3 = instrument_price_factory.create(instrument=product, net_value=300)
        p4 = instrument_price_factory.create(instrument=product, net_value=100)
        assert product.get_high_water_mark(p1.date) == initial_price.net_value
        assert product.get_high_water_mark(p2.date) == p1.net_value
        assert product.get_high_water_mark(p3.date) == p1.net_value
        assert product.get_high_water_mark(p4.date) == p3.net_value
        product.initial_high_water_mark = 400
        product.save()
        assert product.get_high_water_mark(p4.date) == 400



import random
from decimal import Decimal

import pytest
from faker import Faker
from psycopg.types.range import NumericRange
from wbcompliance.factories.risk_management import RuleThresholdFactory

from wbportfolio.risk_management.backends.stop_loss_instrument import (
    RuleBackend as StopLossInstrumentRuleBackend,
)
from wbportfolio.tests.models.utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestStopLossPortfolioRuleModel(PortfolioTestMixin):
    @pytest.fixture
    def stop_loss_portfolio_backend(self, weekday, date_interval_option, freq, product):
        parameters = {"freq": freq, "date_interval_option": date_interval_option}
        lower = random.random()
        upper = random.uniform(lower, 1)
        return StopLossInstrumentRuleBackend(
            weekday,
            product,
            parameters,
            [RuleThresholdFactory.create(range=NumericRange(lower=lower, upper=upper))],  # type: ignore
        )

    @pytest.mark.parametrize(
        "date_interval_option, freq",
        [
            ("ROLLING_WINDOWS", StopLossInstrumentRuleBackend.FreqChoices.BUSINESS_DAY),
            *[("FREQUENCY", option) for option in StopLossInstrumentRuleBackend.FreqChoices.values],
        ],
    )
    def test_check_rule_frequency(
        self,
        weekday,
        date_interval_option,
        freq,
        product,
        portfolio,
        instrument_price_factory,
        asset_position_factory,
        stop_loss_portfolio_backend,
        instrument_portfolio_through_model_factory,
    ):
        instrument_portfolio_through_model_factory.create(
            instrument=product,
            portfolio=portfolio,
            primary_portfolio=True,
        )

        d1 = stop_loss_portfolio_backend._get_start_interval()

        threshold = stop_loss_portfolio_backend.thresholds[0]
        breach_perf = random.uniform(threshold.range.lower, threshold.range.upper)

        i1 = instrument_price_factory.create(date=d1, net_value=100, calculated=False, instrument=product)
        instrument_price_factory.create(
            date=weekday, net_value=Decimal(breach_perf + 1) * i1.net_value, calculated=False, instrument=product
        )

        asset_position_factory.create(date=weekday, underlying_instrument=product, portfolio=portfolio)
        asset_position_factory.create(date=weekday, underlying_instrument=product, portfolio=portfolio)

        res = list(stop_loss_portfolio_backend.check_rule())
        assert len(res) == 1
        assert res[0].breached_object.id == product.id

    @pytest.mark.parametrize(
        "date_interval_option, freq",
        [
            ("ROLLING_WINDOWS", StopLossInstrumentRuleBackend.FreqChoices.BUSINESS_DAY),
            *[("FREQUENCY", option) for option in StopLossInstrumentRuleBackend.FreqChoices.values],
        ],
    )
    def test_check_rule_frequency_2(
        self,
        weekday,
        date_interval_option,
        freq,
        product,
        portfolio,
        instrument_price_factory,
        asset_position_factory,
        instrument_factory,
        stop_loss_portfolio_backend,
        instrument_portfolio_through_model_factory,
    ):
        instrument_portfolio_through_model_factory.create(
            instrument=product,
            portfolio=portfolio,
            primary_portfolio=True,
        )
        d1 = stop_loss_portfolio_backend._get_start_interval()
        benchmark = instrument_factory.create()

        threshold = stop_loss_portfolio_backend.thresholds[0]
        threshold.range = NumericRange(upper=-0.5, lower=None)  # type: ignore
        threshold.save()

        instrument_price_factory.create(date=d1, net_value=100, calculated=False, instrument=product)
        instrument_price_factory.create(date=weekday, net_value=100, calculated=False, instrument=product)

        instrument_price_factory.create(date=d1, net_value=100, calculated=False, instrument=benchmark)
        instrument_price_factory.create(date=weekday, net_value=300, calculated=False, instrument=benchmark)

        asset_position_factory.create(date=weekday, underlying_instrument=product, portfolio=portfolio)
        asset_position_factory.create(date=weekday, underlying_instrument=product, portfolio=portfolio)

        res = list(stop_loss_portfolio_backend.check_rule())
        assert len(res) == 0

        setattr(stop_loss_portfolio_backend, "static_benchmark", benchmark)
        res = list(stop_loss_portfolio_backend.check_rule())
        assert len(res) == 1
        assert res[0].breached_object.id == product.id



from contextlib import suppress
from datetime import date

from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from pandas._libs.tslibs.offsets import BDay
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcompliance.viewsets.risk_management.mixins import RiskCheckViewSetMixin
from wbcore import viewsets
from wbcore.metadata.configs.display.instance_display import (
    Display,
    create_simple_display,
)
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.views import CloneMixin

from wbportfolio.models import AssetPosition, TradeProposal
from wbportfolio.models.transactions.trade_proposals import (
    replay_as_task,
)
from wbportfolio.serializers import (
    ReadOnlyTradeProposalModelSerializer,
    TradeProposalModelSerializer,
    TradeProposalRepresentationSerializer,
)

from ..configs import (
    TradeProposalButtonConfig,
    TradeProposalDisplayConfig,
    TradeProposalEndpointConfig,
    TradeProposalPortfolioEndpointConfig,
)
from ..mixins import UserPortfolioRequestPermissionMixin


class TradeProposalRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:trade"
    queryset = TradeProposal.objects.all()
    serializer_class = TradeProposalRepresentationSerializer


class TradeProposalModelViewSet(CloneMixin, RiskCheckViewSetMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    ordering_fields = ("trade_date",)
    ordering = ("-trade_date",)
    search_fields = ("comment",)
    filterset_fields = {"trade_date": ["exact", "gte", "lte"], "status": ["exact"]}

    queryset = TradeProposal.objects.select_related("rebalancing_model", "portfolio")
    serializer_class = TradeProposalModelSerializer
    display_config_class = TradeProposalDisplayConfig
    button_config_class = TradeProposalButtonConfig
    endpoint_config_class = TradeProposalEndpointConfig

    def get_serializer_class(self):
        if self.new_mode or (
            "pk" in self.kwargs and (obj := self.get_object()) and obj.status == TradeProposal.Status.DRAFT
        ):
            return TradeProposalModelSerializer
        return ReadOnlyTradeProposalModelSerializer

    @classmethod
    def get_clone_instance_display(cls) -> Display:
        return create_simple_display(
            [
                ["comment"],
                ["trade_date"],
            ]
        )

    @classmethod
    def _get_risk_checks_button_title(cls) -> str:
        return "Pre-Trade Checks"

    @action(detail=True, methods=["PATCH"])
    def reset(self, request, pk=None):
        trade_proposal = get_object_or_404(TradeProposal, pk=pk)
        if trade_proposal.status == TradeProposal.Status.DRAFT:
            trade_proposal.reset_trades()
            return Response({"send": True})
        return Response({"status": "Trade proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"])
    def normalize(self, request, pk=None):
        trade_proposal = get_object_or_404(TradeProposal, pk=pk)
        if trade_proposal.status == TradeProposal.Status.DRAFT:
            trade_proposal.normalize_trades()
            return Response({"send": True})
        return Response({"status": "Trade proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"])
    def replay(self, request, pk=None):
        trade_proposal = get_object_or_404(TradeProposal, pk=pk)
        if trade_proposal.portfolio.is_manageable:
            replay_as_task.delay(trade_proposal.id)
            return Response({"send": True})
        return Response({"status": "Trade proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"])
    def deleteall(self, request, pk=None):
        trade_proposal = get_object_or_404(TradeProposal, pk=pk)
        if trade_proposal.status == TradeProposal.Status.DRAFT:
            trade_proposal.trades.all().delete()
            return Response({"send": True})
        return Response({"status": "Trade proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)


class TradeProposalPortfolioModelViewSet(UserPortfolioRequestPermissionMixin, TradeProposalModelViewSet):
    endpoint_config_class = TradeProposalPortfolioEndpointConfig

    @cached_property
    def default_trade_date(self) -> date | None:
        with suppress(AssetPosition.DoesNotExist):
            return (self.portfolio.assets.latest("date").date + BDay(1)).date()

    def get_queryset(self):
        return TradeProposal.objects.filter(portfolio=self.kwargs["portfolio_id"])



# Generated by Django 4.1.11 on 2023-10-04 06:48

import django_fsm
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0047_remove_product_default_sub_account"),
    ]

    operations = [
        migrations.AlterField(
            model_name="trade",
            name="status",
            field=django_fsm.FSMField(
                choices=[
                    ("DRAFT", "Draft"),
                    ("SUBMIT", "Submit"),
                    ("EXECUTED", "Executed"),
                    ("CONFIRMED", "Confirmed"),
                    ("FAILED", "Failed"),
                ],
                default="CONFIRMED",
                max_length=50,
                verbose_name="Status",
            ),
        ),
    ]



import factory

from wbportfolio.models import AccountReconciliation, AccountReconciliationLine


class AccountReconciliationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = AccountReconciliation

    account = factory.SubFactory("wbcrm.factories.AccountFactory")
    creator = factory.SubFactory("wbcore.contrib.authentication.factories.UserFactory")
    reconciliation_date = factory.Faker("date_this_year")


class AccountReconciliationLineFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = AccountReconciliationLine

    reconciliation = factory.SubFactory(AccountReconciliationFactory)
    product = factory.SubFactory("wbportfolio.factories.ProductFactory")
    price = factory.Faker("pydecimal", left_digits=4, right_digits=2)
    price_date = factory.Faker("date_this_year")
    shares = factory.Faker("pyint")
    shares_external = factory.Faker("pyint")



import json
from datetime import datetime
from io import BytesIO
from typing import Optional

from django.db import models
from pandas.tseries.offsets import BDay
from wbcore.contrib.io.backends import AbstractDataBackend, register

from ..utils import process_request
from .mixin import DataBackendMixin


@register("Instrument Prices", provider_key="ubs", save_data_in_import_source=True, passive_only=True)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    def __init__(
        self, import_credential: Optional[models.Model] = None, ubs_bank: Optional[models.Model] = None, **kwargs
    ):
        if not ubs_bank:
            raise ValueError("The ubs company objects needs to be passed to this backend")
        self.ubs_bank = ubs_bank
        if not import_credential or not import_credential.authentication_token:
            raise ValueError("UBS backend needs a valid import credential object")
        self.authentication_token = import_credential.authentication_token

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = (execution_time - BDay(1)).date()
        endpoint = "https://neo.ubs.com/api/ged-amc/external/report/v1/valuation/{0}/{1}"
        if obj_external_ids:
            for external_id in obj_external_ids:
                res_json = process_request(
                    self.authentication_token, endpoint.format(external_id, execution_date.strftime("%Y-%m-%d"))
                )
                res_json.pop("constituents", None)
                if res_json:
                    content_file = BytesIO()
                    content_file.write(json.dumps(res_json).encode())
                    file_name = f"ubs_instrument_price_{external_id}_{execution_date:%Y-%m-%d}_{int(datetime.timestamp(execution_time))}.json"
                    yield file_name, content_file



import random

import factory

from wbportfolio.models import Trade, TradeProposal

from .transactions import TransactionFactory


class TradeFactory(TransactionFactory):
    class Meta:
        model = Trade

    bank = factory.Faker("company")
    marked_for_deletion = False
    shares = factory.Faker("pydecimal", min_value=10, max_value=1000, right_digits=4)
    price = factory.LazyAttribute(lambda o: random.randint(10, 10000))
    # trade_proposal = factory.SubFactory("wbportfolio.factories.TradeProposalFactory")


class TradeProposalFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = TradeProposal

    trade_date = factory.Faker("date_object")
    comment = factory.Faker("paragraph")
    portfolio = factory.SubFactory("wbportfolio.factories.PortfolioFactory")
    creator = factory.SubFactory("wbcore.contrib.directory.factories.PersonFactory")


class CustomerTradeFactory(TradeFactory):
    transaction_subtype = factory.LazyAttribute(
        lambda o: Trade.Type.REDEMPTION if o.shares < 0 else Trade.Type.SUBSCRIPTION
    )
    underlying_instrument = factory.SubFactory("wbportfolio.factories.ProductFactory")
    portfolio = factory.LazyAttribute(lambda x: x.underlying_instrument.primary_portfolio)



from dynamic_preferences.registries import global_preferences_registry


def get_monthly_nnm_target(*args, **kwargs):
    global_preferences = global_preferences_registry.manager()
    return global_preferences["wbportfolio__monthly_nnm_target"]



from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import (
    IntegerPreference,
    LongStringPreference,
    StringPreference,
)

portfolio = Section("wbportfolio")
commission = Section("wbcommission")


@global_preferences_registry.register
class TimedeltaImportInstrumentPrice(IntegerPreference):
    section = portfolio
    name = "timedelta_import_instrument_price"
    default = 2

    verbose_name = "The Timedelta for the instrument price import windows"


@global_preferences_registry.register
class DaysToRecomputeRebateFromFeesThreshold(IntegerPreference):
    section = commission
    name = "days_to_recompute_rebate_from_fees_threshold"
    default = 90  # 1 quarter

    verbose_name = "Number of Days to recompute rebate from fees"


@global_preferences_registry.register
class MonthlyNetNewMoneyTarget(IntegerPreference):
    section = portfolio
    name = "monthly_nnm_target"
    default = int(1e6)

    verbose_name = "Monthly Net New Money Target"


@global_preferences_registry.register
class AccountHoldingReconciliationNotificationTitle(StringPreference):
    section = portfolio
    name = "account_holding_reconciliation_notification_title"
    default = "Account Holdings Reconciliation"


@global_preferences_registry.register
class AccountHoldingReconciliationNotificationBody(LongStringPreference):
    section = portfolio
    name = "account_holding_reconciliation_notification_body"
    default = """To validate your holdings, please review the reconciliation for the account {account} on {reconciliation_date}."""


@global_preferences_registry.register
class AccountHoldingReconciliationNotificationBodyUpdate(LongStringPreference):
    section = portfolio
    name = "account_holding_reconciliation_notification_body_update"
    default = """A reconcilation has been updated and requires your review. Please review the reconciliation for the account {account} on {reconciliation_date}."""
